/* ACG_CG.C -- Assembly Constraint Graph Consequence Generation routines */
/* (c) 1988 - University of Michigan */

#include <stdio.h>
#include "config.h"
#include "global.h"
#include "set.h"
#include "assert.h"
#include "acg.h"

/* CONSEQUENCE(leaf,proposal)
 * Given a proposed sequencing or trajectory proposal, find all consequences
 * by calling crit_sequ() or crit_traj() for each one.  Return TRUE if
 * the proposal leads to a contradiction.
 */

boolean consequence(leaf,prop)
Leaf leaf;
Prop *prop;
{
SequProp *sp;
TrajProp *tp;

	if (prop->type == SEQU)
	{
		sp = prop->a.sequ;
		if (isin(&leaf[DATA]->graf[sp->p2].pred, sp->p1))
			error("redundant sequencing proposal\n");
		return(sequ_consequence(leaf,sp));
	}
	else
	{
		tp = prop->a.traj;
		if (isequal(&tp->t, &leaf[DATA]->graf[tp->p].traj))
			error("redundant trajectory proposal\n");
		return(traj_consequence(leaf,tp));
	}
}

/* SEQU_CONSEQUENCE(leaf,proposal)
 * Find all consequences of a sequencing proposal by calling sequ_assert()
 * or traj_assert() for each one.  Return TRUE if the proposal is
 * leads to an contradiction.
 */

boolean sequ_consequence(leaf,sp)
Leaf leaf;
SequProp *sp;
{
register Part p1, p2;
int n;
SequAsrt sa;
TrajProp tp;
Graf *graf = leaf[DATA]->graf;
Constraint *con, *ocon;

	if (is_on(SHOW_ENTRIES))
	{
		printf("sequ_consequence ");
		dump_sequprop(stdout,sp);
	}

	/* Check if p1 => p2 is a legal proposal */
	if ((p1 = sp->p1) == (p2 = sp->p2))
		error("part proposed to preceed itself\n");
	if (isin(&graf[p1].pred,p2))
		return(TRUE);

	/* Check if it is a new proposal */
	if (isin(&graf[p1].succ,p2))
		return(FALSE);

	/* Tell all the criteria about it */
	sa.p1 = &graf[p1].pred;
	sa.p2 = &graf[p2].succ;
	crit_sequ(leaf,&sa);

	/* Update the sequence graph */
	foreach(p1,sa.p1)
	{
		append(&graf[p1].succ,sa.p2);
		if ((n = cardinality(&graf[p1].succ)) != graf[p1].nsucc)
		{
			graf[p1].nsucc = n;
			if (n + graf[p1].npred == part_number() + 1)
				leaf[DATA]->sequ_inc--;
		}
	}
	foreach(p2,sa.p2)
	{
		append(&graf[p2].pred,sa.p1);
		if ((n = cardinality(&graf[p2].pred)) != graf[p2].npred)
		{
			graf[p2].npred = n;
			if (graf[p2].nsucc + n == part_number() + 1)
				leaf[DATA]->sequ_inc--;
		}
	}

#ifdef PRIVCON
	/* Look for a constraint implying p1 => p2.  Drop all these. */
	foreach(p1,sa.p1)
	{
		con = constraint1(graf,p1);
		ocon = NULL;
		while (con != NULL)
		{
			if (isin(sa.p2,con->part))
				/* Get rid of the constraint */
				if (ocon)
				{
					ocon->next = con->next;
					drop_constraint(con);
					con = ocon->next;
				}
				else
				{
					constraint1(graf,p1) = con->next;
					drop_constraint(con);
					con = constraint1(graf,p1);
				}
			}
			else
			     con = (ocon=con)->next;
	}
#endif PRIVCON

	/* Look for a trajectory implying p2 => p1.  Assert the negation. */
	foreach(p2,sa.p2)
	{
		con = constraint1(graf,p2);
		ocon = NULL;
		while (con != NULL)
		{
			if (isin(sa.p1,con->part))
			{
				/* Assert the negation of the trajectory */
				setdifference(&tp.t,
					      &graf[tp.p=p2].traj,
					      &con->traj);
				if (isempty(&tp.t)) return(TRUE);
#ifdef PRIVCON
				/* Get rid of the constraint */
				if (ocon)
				{
					ocon->next = con->next;
					drop_constraint(con);
					con = ocon->next;
				}
				else
				{
					constraint1(graf,p2) = con->next;
					drop_constraint(con);
					con = constraint1(graf,p2);
				}
#endif PRIVCON
				if (traj_consequence(leaf,&tp))
					return(TRUE);
			}
#ifdef PRIVCON
			else
#endif PRIVCON
				 con = (ocon=con)->next;
		}
	}
	return(FALSE);
}

/* TRAJ_CONSEQUENCE(leaf,proposal)
 * Find all consequences of a trajectory proposal by calling crit_sequ()
 * or crit_traj() for each one. Return TRUE if the proposal leads to a
 * contradiction.
 */

boolean traj_consequence(leaf,tp)
Leaf leaf;
TrajProp *tp;
{
Graf *graf = leaf[DATA]->graf;
register TrajSet *newtraj = &tp->t;
register TrajSet *oldtraj = &graf[tp->p].traj;
Constraint *con, *ocon, *tcon;
SequProp sp;
PartSet tps;

	if (is_on(SHOW_ENTRIES))
	{
		printf("traj_consequence ");
		dump_trajprop(stdout,tp);
	}

	/* Check if this is a legal proposal */
	if (!issubset(newtraj,oldtraj))
		error("proposed trajectory set is not a subset\n");
	if (isempty(newtraj))
		error("proposed trajectory set is empty\n");

	/* Check if it is a new proposal */
	if (isequal(newtraj,oldtraj))
		return(FALSE);

	/* Tell all the criteria about it */
	crit_traj(leaf,tp);

	/* Update the trajectory lists */
	setcopy(oldtraj,newtraj);
	/* Check if this node is done now */
	if ((graf[tp->p].ntraj = cardinality(newtraj)) == 1)
		leaf[DATA]->traj_inc--;

	/* Look for constraints that are fired.  Assert their consequences */
	con = constraint1(graf,tp->p);
	ocon=NULL;
	while (con != NULL)
	{
#ifdef PRIVCON
		if (!intersect(&con->traj,newtraj))
		{
			if (ocon)
				tcon = ocon->next = con->next;
			else
				tcon = constraint1(graf,tp->p) = con->next;
			drop_constraint(con);
			con = tcon;
		}
		else
#endif PRIVCON
		if (issubset(newtraj,&con->traj))
		{
			/* Assert the consequence */
			sp.p1 = tp->p;
			sp.p2 = con->part;
#ifdef PRIVCON
			if (ocon)
				ocon->next = con->next;
			else
				constraint1(graf,tp->p) = con->next;
			drop_constraint(con);
#else
			con = con->next;
#endif PRIVCON
			if (sequ_consequence(leaf,&sp))
				return(TRUE);
#ifdef PRIVCON
			/* The call to sequ_consequence may have deallocated */
			/* parts of the constraint list, so it's not safe to */
			/* try to continue from where we were.  So we rewind */
			/* to the beginning, which may not be the best thing */
			/* to do about it, but is probably the simplest.     */
			con = constraint1(graf,tp->p);
			ocon = NULL;
#endif
		}
		else
			con = (ocon=con)->next;
	}

	return(FALSE);
}

/* IMMEDIATE_CONSEQUENCE(leaf)
 * Find all immediate consequences of the initial.  This is used on the
 * initial data structure to see fire any constraints that are already
 * applicable.  Returns TRUE if a conflict arises.
 */

boolean immediate_consequence(leaf)
Leaf leaf;
{
Graf *graf = leaf[DATA]->graf;
Constraint *con, *ocon, *tcon;
SequProp sp;
Part p;

    foreachpart(p)
    {
	/* Look for constraints that are fired.  Assert their consequences */
	con = constraint1(graf,p);
	ocon=NULL;
	while (con != NULL)
	{
		if (issubset(&graf[p].traj,&con->traj))
		{
			/* Assert the consequence */
			sp.p1 = p;
			sp.p2 = con->part;
#ifdef PRIVCON
			if (ocon)
				ocon->next = con->next;
			else
				constraint1(graf,p) = con->next;
			drop_constraint(con);
#else
			con = con->next;
#endif PRIVCON
			if (sequ_consequence(leaf,&sp))
				return(TRUE);
#ifdef PRIVCON
			/* The call to sequ_consequence may have deallocated */
			/* parts of the constraint list, so it's not safe to */
			/* try to continue from where we were.  So we rewind */
			/* to the beginning, which may not be the best thing */
			/* to do about it, but is probably the simplest.     */
			con = constraint1(graf,p);
			ocon = NULL;
#endif
		}
		else
			con = (ocon=con)->next;
	}
    }

	return(FALSE);
}
