/* ACG_INIT.C  -- Init routines for Assembly Constraint Graph Module */
/* (c) 1988 - University of Michigan */ 

#include <stdio.h>
#include "config.h"
#include "global.h"
#include "set.h"
#include "assert.h"
#include "acg.h"

#ifdef PRIVCON
char *constraint_mm;
#else
Constraint **constraint_list;
#endif PRIVCON

/* ACG_INIT(leaf)
 * This routine creates an initial acg leaf by querying the cad interface
 * for data.  Somebody should have selected a cad file to read from before
 * this got called.  It also initializes the static acg structure used for
 * the constraint list.
 */

acg_init(leaf)
Leaf leaf;
{
int parts;
Part p,q;
Traj t;
PartSet *ps;
Constraint *c;
Graf *graf;

	if (is_on(SHOW_ENTRIES)) printf("acg_init\n");
	parts = part_number();

	/* Get memory for the leaf */
	leaf[DATA] = (struct Leaf_Data *) m_alloc(sizeof(struct Leaf_Data));
	leaf[DATA]->graf = graf = (Graf *) m_alloc(parts*sizeof(Graf));
	leaf[DATA]->sequ_inc = (parts > 1) ? parts:0;
	leaf[DATA]->traj_inc = 0;

#ifdef PRIVCON
	init_priv_con();
#else
	constraint_list = (Constraint **) m_alloc(parts*sizeof(Constraint *));
#endif

	/* Load up the possible trajectories and constraints */
	for (p=0; p < parts; p++)
	{
		/*
		 * Initialize each part to be only it's own predecessor and
		 * successor.  Get it's proposed trajectories from cad.
		 */
		setpartset(&graf[p].pred,p);
		setpartset(&graf[p].succ,p);
		graf[p].npred = 1;
		graf[p].nsucc = 1;

		setcopy(&graf[p].traj,proposed_traj(p));
		if ((graf[p].ntraj = cardinality(&graf[p].traj)) > 1)
			leaf[DATA]->traj_inc++;

		/* Set up the Constraint List */

		constraint1(graf,p) = NULL;
		foreach(t,(&graf[p].traj))
		{
		    /* Get set of parts which must follow p if t is used */
		    if (ps = constrained_parts(p,t))
		    {
			/* merge as many as possible into old constraints */
			for (c=constraint1(graf,p); c != NULL; c=c->next)
				if (isin(ps,c->part))
				{
					append1((&c->traj),t);
					remove1(ps,c->part);
				}
			/* For any other parts, create new constraints */
			foreach (q,ps)
			{
				c = alloc_constraint();
				c->next = constraint1(graf,p);
				constraint1(graf,p) = c;
				settrajset(&c->traj,t);
				c->part = q;
			}
		    }
		}
	}
}

/* ACG_COPY(new,old)
 * This creates a new acg data structure which is a copy of the one in the old
 * leaf and attaches it to the new leaf.
 */

acg_copy(new,old)
Leaf new,old;
{
register Graf *ng, *og;
Part p;
int parts = part_number();
#ifdef PRIVCON
Constraint *oc, *nc, *ncp;
#endif

	if (is_on(SHOW_ENTRIES)) printf("acg_copy\n");

	/* Get memory for the frame */
	new[DATA] = (struct Leaf_Data *) m_alloc(sizeof(struct Leaf_Data));
	new[DATA]->traj_inc = old[DATA]->traj_inc;
	new[DATA]->sequ_inc = old[DATA]->sequ_inc;

	ng = new[DATA]->graf = (Graf *) m_alloc(parts*sizeof(Graf));
	og = old[DATA]->graf;

	/* Load up the graf and possible trajectories */
	foreachpart(p)
	{
		setcopy(&ng[p].pred,&og[p].pred);
		setcopy(&ng[p].succ,&og[p].succ);
		setcopy(&ng[p].traj,&og[p].traj);
		ng[p].npred = og[p].npred;
		ng[p].nsucc = og[p].nsucc;
		ng[p].ntraj = og[p].ntraj;
#ifdef PRIVCON
		/* Copy over constraints, if they are private */
		constraint1(ng,p) = ncp = NULL;
		for (oc=constraint1(og,p); oc != NULL; oc = oc->next)
		{
			nc = alloc_constraint();
			if (ncp)
				ncp->next = nc;
			else
				constraint1(ng,p) = nc;
			setcopy(&nc->traj,&oc->traj);
			nc->part = oc->part;
			(ncp=nc)->next = NULL;
		}
#endif PRIVCON
	}
}

/* ACG_DROP(leaf)
 * Dispose of the data structure for the given leaf.
 */

acg_drop(leaf)
Leaf leaf;
{
int parts = part_number();
#ifdef PRIVCON
Part p;
Constraint *c, *nc;
#endif PRIVCON

	if (is_on(SHOW_ENTRIES)) printf("acg_drop\n");
#ifdef NORMAL
	drop_norm();
#endif NORMAL
#ifdef PRIVCON
	foreachpart(p)
		for (c=constraint1(leaf[DATA]->graf,p); c != NULL; c = nc)
		{
			nc = c->next;
			drop_constraint(c);
		}
#endif PRIVCON
	m_free(leaf[DATA]->graf,parts*sizeof(Graf));
	m_free(leaf[DATA],sizeof(struct Leaf_Data));
	leaf[DATA] = NULL;
}

/* ACG_RATE(leaf)
 * Rate a leaf by how near it is to completion.
 */

int acg_rate(leaf)
Leaf leaf;
{
#ifdef COMP1
	/* Return 1 if incomplete, 0 if complete */
	if (is_on(SHOW_ENTRIES)) printf("acg_rate\n");
	return(leaf[DATA]->traj_inc + leaf[DATA]->sequ_inc != 0);
#else
#ifdef COMP2
	/* Return number of completed parts */
	if (is_on(SHOW_ENTRIES)) printf("acg_rate\n");
	return(leaf[DATA]->traj_inc + leaf[DATA]->sequ_inc);
#else
register int s,t;
register Part p;
int parts = part_number();
register Graf *graf = leaf[DATA]->graf;

	if (is_on(SHOW_ENTRIES)) printf("acg_rate\n");
	/* Add up number of assertions not yet made */
	s=t=0;
	for (p=0; p < parts; p++)
	{
		s += graf[p].npred + graf[p].nsucc;
		t += graf[p].ntraj;
	}
		
	return(parts*parts - s + t);
#endif COMP2
#endif COMP1
}

#ifdef SIZE
long acg_leaf_size(leaf)
Leaf leaf;
{
int parts = part_number();
long size;
#ifdef PRIVCON
Part p;
Constraint *c;
#endif PRIVCON

	size = parts*sizeof(Graf) + sizeof(struct Leaf_Data);

#ifdef PRIVCON
	foreachpart(p)
		for (c=constraint1(leaf[DATA]->graf,p); c != NULL; c = c->next)
			size += sizeof(Constraint);
#endif PRIVCON
	return(size);
}

long acg_stat_size()
{
long size;
Part p;
Constraint *c;

#ifndef PRIVCON
	/* Memory for constraint list */
	size = part_number()*sizeof(Constraint *);

	foreachpart(p)
		for (c=constraint1(NULL,p); c != NULL; c = c->next)
			size += sizeof(Constraint);
#else
	/* Memory wasted by memory manager */
	size = mm_size(constraint_mm);
#endif PRIVCON
	return(size);
}
#endif SIZE
