#include <stdio.h>
#include "config.h"
#include "set.h"
#include "global.h"
#include "assert.h"

#define DATA BIG_MOD

/* Structure for Leaf_Data */

struct Leaf_Data {
#ifdef MERGE1
	int link_cnt;           /* How many leaves are using this */
#endif MERGE1

	short *seqn;		/* sequence #s for parts, NULL if unknown */
	int rating;		/* Current rating of leaf, -1 if unknown */
	int sprate;		/* Rating of best proposal, -2 if unknown */
	SequProp sprop;		/* Best sequencing proposal */
#ifdef RETIRE
	boolean retired;	/* have we retired? */
#endif RETIRE
	};

/* Static Data Structure */

int manip[MAXPARTS];		/* Manipulabilities */
Part psort[MAXPARTS];		/* Sorted list of parts by manipulabilities */


int pcmp(a,b)
Part *a,*b;
{
	return (manip[*b] - manip[*a]);
}

/* BIG_INIT(leaf)
 * Initialize the data structure for the root leaf given, by allocating memory.
 */

big_init(leaf)
Leaf leaf;
{
int parts = part_number();
Part p;

	if (is_on(SHOW_ENTRIES)) printf("big_init\n");

	/* Initialize static structures */

	for (p=0; p<parts; p++)
	{
		manip[p] = get_mani(p);
		psort[p] = p;
	}
	qsort(psort,parts,sizeof(Part),pcmp);

	/* Initialize root leaf */

	leaf[DATA] = (struct Leaf_Data *) m_alloc(sizeof(struct Leaf_Data));

	leaf[DATA]->seqn = NULL;	/* Unknown */
	leaf[DATA]->rating = -1;	/* Unknown */
	leaf[DATA]->sprate = -2;	/* Unknown */
#ifdef RETIRE
	leaf[DATA]->retired = FALSE;
#endif RETIRE
#ifdef MERGE1
	leaf[DATA]->link_cnt = 1;
#endif MERGE1

}


/* BIG_COPY(new,old)
 * Copy the given leaf into a new one.  This is done by making both point to
 * the same leaf if MERGE1 is defined.  Otherwise a physical copy is always
 * made.
 *
 * BIG_SPLIT(leaf)
 * This takes a leaf with a link count greater than one and divides it into a
 * separate leaf with it's own memory.  Used if MERGE1 is defined only.
 */

#ifdef MERGE1
big_copy(new,old)
Leaf new,old;
{
	if (is_on(SHOW_ENTRIES)) printf("big_copy\n");

	new[DATA] = old[DATA];
	new[DATA]->link_cnt++;
}

big_split(new)
Leaf new;
{
struct Leaf_Data *od = new[DATA];
#else
big_copy(new,old)
Leaf new,old;
{
struct Leaf_Data *od = old[DATA];
#endif
int parts = part_number();
Part p;

	if (is_on(SHOW_ENTRIES)) printf("big_split\n");

	/* Allocate memory for the new leaf */
	new[DATA] = (struct Leaf_Data *) m_alloc(sizeof(struct Leaf_Data));

#ifdef MERGE1
	new[DATA]->link_cnt = 1;
	od->link_cnt--;
#endif MERGE1
#ifdef RETIRE
	new[DATA]->retired = od->retired;
#endif RETIRE

	/* Don't copy the sequence numbers, as they are won't be useful.   */
	/* We wouldn't be splitting this node if we weren't going to redo  */
	/* the sequence numbers and rating.				   */

	new[DATA]->seqn = NULL;
	new[DATA]->rating = -1;
}


/* BIG_DROP(leaf)
 * Release all memory used by our data structure in the given leaf.
 */

big_drop(leaf)
Leaf leaf;
{
Part p;
int parts = part_number();

	if (is_on(SHOW_ENTRIES)) printf("big_drop\n");

#ifdef MERGE1
	if (--leaf[DATA]->link_cnt == 0)
#endif MERGE1
	{
		big_retire(leaf);
		m_free(leaf[DATA],sizeof(struct Leaf_Data));
	}
}

big_retire(leaf)
Leaf leaf;
{
Part p;
int parts = part_number();

	if (is_on(SHOW_ENTRIES)) printf("big_retire\n");

	if (leaf[DATA]->seqn != NULL)
	{
		m_free(leaf[DATA]->seqn,sizeof(short)*parts);
		leaf[DATA]->seqn = NULL;
	}
#ifdef RETIRE
	leaf[DATA]->retired = TRUE;
#endif RETIRE
}

/* BIG_DUMP(leaf,fp)
 * print the contents of the our leaf data on the given device.
 */

big_dump(leaf,fp)
Leaf leaf;
FILE *fp;
{
Part p;
int parts = part_number();
short *sn = leaf[DATA]->seqn;

#ifdef RETIRE
	if (leaf[DATA]->retired)
#ifdef MERGE1
	    fprintf(fp,"BIG Retired. rating=%d. (lc=%d)\n",
		leaf[DATA]->link_cnt,leaf[DATA]->rating);
#else
	    fprintf(fp,"BIG Retired. rating=%d.\n",leaf[DATA]->rating);
#endif MERGE1
        else
#endif RETIRE
	{
#ifdef MERGE1
	    fprintf(fp,"BIG. (lc=%d) rating=%d.\n",
		leaf[DATA]->link_cnt,leaf[DATA]->rating);
#else
	    fprintf(fp,"BIG. rating=%d.\n",leaf[DATA]->rating);
#endif MERGE1
	    if (sn == NULL)
		fprintf(fp,"  No Sequence Numbers\n");
	    else
		for (p=0; p<parts;p++)
		    fprintf(fp,"  %2d: Part %s\n",p,part_name(sn[p]));
	}
}


/* BIG_TRAJ(leaf,ta)
 * BIG_SEQU(leaf,sa)
 * Update the data structure to account for the given assertions.
 */

big_traj(leaf,ta)
Leaf leaf;
TrajAsrt *ta;
{
	/* Do Nothing */
}

big_sequ(leaf,sa)
Leaf leaf;
SequAsrt *sa;
{
register Part p,q;
short *sn = leaf[DATA]->seqn;

    if (is_on(SHOW_ENTRIES)) printf("big_sequ\n");
 
#ifdef RETIRE
    /* If retired, ignore updates */
    if (leaf[DATA]->retired) return;
#endif RETIRE
    /* If no stored sequence numbers, ignore updates */
    if (leaf[DATA]->rating == -1) return;

    /* If the assertion contradicts the stored sequence, throw numbers out */
    foreach (p,sa->p1)
	foreach (q,sa->p2)
	    if (sn[p] > sn[q])
	    {
#ifdef MERGE1
		if (leaf[DATA]->link_cnt > 1)
		    big_split(leaf);
#endif MERGE1
		leaf[DATA]->rating = -1;
	    }
}


/* BIG_RATE(leaf)
 * Make an optimistic estimate of the rating of this plan.
 */

int big_rate(leaf)
Leaf leaf;
{
short *sn=leaf[DATA]->seqn;
int parts=part_number();
Part p,q;
int seq;
int rate;

	if (is_on(SHOW_ENTRIES)) printf("big_rate\n");

	if (leaf[DATA]->rating != -1)
		return(leaf[DATA]->rating);

	/* Any old proposal is no longer valid...discard it */
	leaf[DATA]->sprate = -2;

	/* Allocate memory for sequence numbers, if we haven't got any */
	if (sn == NULL)
	    sn = leaf[DATA]->seqn = (short *) m_alloc(sizeof(short)*parts);

	/* Compute new sequence numbers */
	rate = 0;
	foreachpart(p)
	    sn[p] = -1;

	foreachpart(q)
	{
		p = psort[q];
		for (seq = n_pred_parts(leaf,p); sn[seq] != -1; seq++)
			if (seq >= parts)
				error("Cannot find sequence\n");
		sn[seq] = p;
		rate += manip[p]*seq;
	}

	return (leaf[DATA]->rating = rate);
}


/* BIG_PROP(leaf,prop)
 * Set prop to a proposal that would be bad for big.  Return an estimate of how
 * much this proposal would hurt the rating.  If we can't find a proposal that
 * would make a difference, we return -1.
 */

int big_prop(leaf,prop)
Leaf leaf;
Prop *prop;
{
register Part p,q;
register Part xp,xq;
int parts = part_number();
int best;
int diff;
static SequProp sp;

	if (is_on(SHOW_ENTRIES)) printf("big_prop\n");

#ifdef RETIRE
	/* If we are retired, just pass */
	if (leaf[DATA]->retired) return(-1);
#endif RETIRE

	prop->type = SEQU;
	prop->a.sequ = &sp;

	/* If we still have a favorite proposal sitting around, use it */
	if (leaf[DATA]->sprate != -2)
	{
		sp.p1 = leaf[DATA]->sprop.p1;
		sp.p2 = leaf[DATA]->sprop.p2;
		return(leaf[DATA]->sprate);
	}

	/* Build up a list of potential proposals and rate them */
	best = 0;
	for (xp=0;xp<parts;xp++)
	{
	    p = psort[xp];
	    for (xq=xp+1;xq<parts;xq++)
	    {
		q = psort[xq];
		if (!is_succ(leaf,p,q) && (diff= manip[p] - manip[q]) > best)
		{
			best = diff;
			sp.p1 = p;
			sp.p2 = q;
		}
	    }
	}

	if (best == 0)
	{
#ifdef RETIRE
		big_retire(leaf);
#endif RETIRE
		return(-1);
	}

	leaf[DATA]->sprop.p1 = sp.p1;
	leaf[DATA]->sprop.p2 = sp.p2;
	leaf[DATA]->sprate = best;
	return(best);
}
