/* CAD_READER.C -- routines to read in data for fake cad interface */
/* (c) 1988 - University of Michigan */

#include <stdio.h>
#include "config.h"
#include "set.h"
#include "cad.h"

/* Numbers of Things - assume large until proven otherwise */

int n_parts = MAXPARTS;
int n_directions = MAXPARTS;
int n_trajectories = MAXPARTS;

/* Name Tables
 * Names for Parts, Trajectories, and Directions are stored in the following
 * data structures, which are indexed by their numbers.  The goofiness with
 * the lengths is to ensure that they are even, so as not to confuse Sun's
 * dbx, which has problems with odds row sizes in character arrays.
 */

char p_name[MAXPARTS][(MAX_NAME_LENGTH+2) & ~1];
char d_name[MAXTRAJS][(MAX_NAME_LENGTH+2) & ~1];
char t_name[MAXTRAJS][(MAX_NAME_LENGTH+2) & ~1];

/* Set of trajectories proposed for a given part */

TrajSet *proposed[MAXPARTS];

/* Data Structures for information destined for SFC Criteria */

PartSet *attach[MAXPARTS], *lift[MAXPARTS];
struct rset *rotate[MAXPARTS];
#ifdef CONTACT
PartSet *contact[MAXPARTS];
#endif CONTACT

/* Data Structures for information destined for Manipulability Criteria */

int manip[MAXPARTS];
int diffic[MAXTRAJS];

/* Data Structures for information destined for Monodirectionality Criteria */

TrajSet *dir_traj[MAXTRAJS];

/* List of constraints for each part */

struct con *constraint[MAXPARTS];

/* Keyword Table
 * The following are the keywords for files and their numbers.
 */

char *keyword[] = {
	"PARTS",
#define	PARTS		0
	"TRAJECTORIES",
#define	TRAJECTORIES	1
	"DIRECTIONS",
#define	DIRECTIONS	2
	"PART",
#define	PART		3
	"TRAJECTORY",
#define	TRAJECTORY	4
	"DIRECTION",
#define	DIRECTION	5
	"PROPOSE",
#define	PROPOSE		6
	"BEFORE",
#define	BEFORE		7
	"ATTACH",
#define	ATTACH		8
	"LIFT",
#define	LIFT		9
	"ROTATE",
#define	ROTATE		10
	"CONTACT",
#define	CONTAC		11
	"MANIPULABILITY",
#define	MANIPULABILITY	12
	"DIFFICULTY"
#define	DIFFICULTY	13
	};
#define N_KEYWORDS	14

/* Declarations for local functions */
char *next_word();
int skip_get(),read_keyword(),read_direction(),read_trajectory(),read_part();
int read_val();
#define read_value(fp) read_val(fp,TRUE);

/* READFILE
 * This reads in a data file from which to answer future queries about the
 * assembly.  In real life, most of this information would come from a CAD
 * system, but I have no particular interest in interfacing this to one
 * just now.
 */

init_cad(fp)
FILE *fp;
{
char *word;
int i,bf,k;
Part p,q;
Traj t;
Dire d;
struct con *c1;
struct rset *rsp;

	if (is_on(SHOW_ENTRIES)) printf("init_cad\n");
	/* Read part declarations */
	if (read_keyword(fp) != PARTS)
		error("Keyword PARTS not found\n");
	n_parts=0;
	while ((word = next_word(fp)) > (char *)1)
	{
		if (n_parts == MAXPARTS)
			error("More than %d parts declared\n",MAXPARTS);
		for (i = 0; i < n_parts; i++)
			if (!strcmp(word,p_name[i]))
				error("Duplicate part name: %s\n",word);
		strcpy(p_name[n_parts++],word);
	}
	if (word == NULL) error ("Premature End Of File\n");

	/* Read trajectory declarations */
	if (read_keyword(fp) != TRAJECTORIES)
		error("Keyword TRAJECTORIES not found\n");
	n_trajectories = 0;
	while ((word = next_word(fp)) > (char *)1)
	{
		if (n_trajectories == MAXTRAJS)
			error("More than %d trajectories declared\n",MAXTRAJS);
		for (i = 0; i < n_trajectories; i++)
			if (!strcmp(word,t_name[i]))
				error("Duplicate trajectory name: %s\n",word);
		strcpy(t_name[n_trajectories++],word);
	}
	if (word == NULL) error ("Premature End Of File\n");

	/* Read direction declarations */
	if (read_keyword(fp) != DIRECTIONS)
		error("Keyword DIRECTIONS not found\n");
	n_directions = 0;
	while ((word = next_word(fp)) > (char *)1)
	{
		if (n_directions == MAXTRAJS)
			error("More than %d directions declared\n",MAXTRAJS);
		for (i = 0; i < n_directions; i++)
			if (!strcmp(word,d_name[i]))
				error("Duplicate direction name: %s\n",word);
		strcpy(d_name[n_directions++],word);
	}
	if (word == NULL) error ("Premature End Of File\n");

	/* initialize some empty sets */
	for (p = 0; p < n_parts; p++)
	{
		attach[p] = mkpartset();
		lift[p] = mkpartset();
#ifdef CONTACT
		contact[p] = mkpartset();
#endif CONTACT
		proposed[p] = mktrajset();
		rotate[p] = NULL;
		constraint[p] = NULL;
	}
	for (d = 0; d < n_directions; d++)
		dir_traj[d] = mktrajset();

	/* read part data */
	if (read_keyword(fp) == PART)
	   for (;;)
	   {
		if ((p = read_part(fp)) < 0) error("Missing PART name\n");

		/* read data lumps */
		bf = FALSE;
		for (;;)
		    switch(k=read_keyword(fp))
		    {
		    case ATTACH:
			while ((i = read_part(fp)) >= 0)
				append1(attach[p],i);
			bf = FALSE;
			break;
		    case LIFT:
			while ((i = read_part(fp)) >= 0)
				append1(lift[p],i);
			bf = FALSE;
			break;
		    case CONTAC:
#ifndef CONTACT
			while ((i = read_part(fp)) >= 0)
				;
#else
			while ((i = read_part(fp)) >= 0)
				append1(contact[p],i);
#endif CONTACT
			bf = FALSE;
			break;
		    case ROTATE:
			rsp = (struct rset *) m_alloc(sizeof(struct rset));
			emptypartset(&rsp->p);
			rsp->next = rotate[p];
			rotate[p] = rsp;
			while ((i = read_part(fp)) >= 0)
				append1(&rsp->p,i);
			bf = FALSE;
			break;
		    case MANIPULABILITY:
			manip[p] = read_value(fp);
			bf = FALSE;
			break;
		    case PROPOSE:
			if ((t=read_trajectory(fp)) < 0)
			    error("Missing PROPOSE trajectory for part %s\n",
			    	p_name[p]);
			append1(proposed[p],t);
			bf = TRUE;
			break;
		    case BEFORE:
			if (!bf) error ("BEFORE must follow PROPOSE\n");

			c1 = (struct con *)m_alloc(sizeof(struct con));
			c1->next = constraint[p];
			constraint[p] = c1;

			c1->tr = t;
			c1->sp = mkpartset();

			while ((q=read_part(fp)) >= 0)
				append1(c1->sp,q);
			break;
		    case PART:
			goto next_part;
		    case TRAJECTORY:
			goto do_traj;
		    default:
			error("Unexpected Keyword %s in PART %s block\n",
				keyword[k],p_name[p]);
		    }
next_part:;
	    }

do_traj:
	for (;;)
	{
		if ((t = read_trajectory(fp)) < 0)
			error("Missing TRAJECTORY name\n");

		/* read data lumps */
		for (;;)
		    switch(k=read_keyword(fp))
		    {
		    case DIRECTION:
			if ((d=read_direction(fp)) < 0)
				error ("Missing DIRECTION name\n");
			append1(dir_traj[d],t);
			break;
		    case DIFFICULTY:
			diffic[t] = read_value(fp);
			break;
		    case TRAJECTORY:
			goto next_traj;
		    case -1:
			return;
		    default:
			error("Unexpected Keyword %s in TRAJECTORY %s block\n",
				keyword[k],t_name[t]);
		    }
next_traj:;
	}
}


/* READ VAL(fp,err)
 * Read in an integer.  Anything else is an if err is true.  If err is false,
 * return -1 if there is a newline instead of a value.
 */

int read_val(fp,err)
FILE *fp;
{
int ch, n;

	if ((ch=skipgetc(fp)) < '0' || ch > '9')
	{
		if (err || ch != '\n')
			error("Missing numeric argument\n");
		else
			return(-1);
	}
	n = ch - '0';
	while ((ch=fgetc(fp)) >= '0' && ch <= '9')
		n = 10*n + ch - '0';
	ungetc(ch,fp);
	return(n);
}

/* READ_PART(fp)
 * READ_TRAJECTORY(fp)
 * READ_DIRECTION(fp)
 * Read the name of a whatever and return it's index.  -1 is returned if a
 * newline is found.  Things blow up if an EOF is found, or if the name is
 * not found.
 */

int read_part(fp)
FILE *fp;
{
char *word;
register int i;

	if ((word = next_word(fp)) == NULL)
		error("Premature End of File\n");
	
	if (word  == (char *)1) return(-1);

	for (i=0; i<n_parts; i++)
		if (!strcmp(word, p_name[i]))
			return(i);

	error("Undefined Part Name: %s\n",word);
}

int read_trajectory(fp)
FILE *fp;
{
char *word;
register int i;

	if ((word = next_word(fp)) == NULL)
		error("Premature End of File\n");
	
	if (word  == (char *)1) return(-1);

	for (i=0; i<n_trajectories; i++)
		if (!strcmp(word, t_name[i]))
			return(i);

	error("Undefined Trajectory Name: %s\n",word);
}

int read_direction(fp)
FILE *fp;
{
char *word;
register int i;

	if ((word = next_word(fp)) == NULL)
		error("Premature End of File\n");
	
	if (word  == (char *)1) return(-1);

	for (i=0; i<n_directions; i++)
		if (!strcmp(word, d_name[i]))
			return(i);

	error("Undefined Direction Name: %s\n",word);
}

/* READ_KEYWORD(fp)
 * This returns the numeric code for the next keyword.  Newlines and spaces
 * are skipped.  -1 is returned on EOF.
 */

int read_keyword(fp)
FILE *fp;
{
char *word;
register int i;

	/* Get a word--skipping newlines */
	while ((word = next_word(fp)) == (char *)1)
		;

	/* If EOF, return it */
	if (word == NULL) return(-1);

	for (i=0; i<N_KEYWORDS; i++)
		if (!strcmp(word, keyword[i]))
			return(i);

	error("Bad Keyword: %s\n",word);
}

/* NEXTWORD
 * Return the next word (delimited by spaces or tabs) from the file.  A
 * 1 is returned if a newline is encountered.  More than one newline
 * in a row is treated as a single newline.  A NULL is returned on EOF.
 */

char *next_word(fp)
FILE *fp;
{
register int ch,i;
static char buf[MAX_NAME_LENGTH];

	if ((ch = skipgetc(fp)) == '\n')
	{
		/* Skip blank lines*/
		while ((ch = skipgetc(fp)) == '\n')
			;
		ungetc(ch,fp);
		return((char *)1);
	}

	/* Read word until space, tab, or newline */
	for (i=0; i<MAX_NAME_LENGTH; i++)
	{
		if (ch == EOF)
			return(NULL);
		buf[i] = ch;
		if ((ch = fgetc(fp)) == ' ' || ch == '\t' || ch == '\n')
			goto done;
	}

	/* If ran out of space in buffer, skip rest of word */
	while ((ch = fgetc(fp)) != ' ' && ch != '\t')
		;
done:
	ungetc(ch,fp);
	buf[i+1] = '\0';
	return(buf);
}


/* SKIPGETC(fp)
 * Read past an arbitrary number of spaces and tabs, and return the first non-
 * blank character.  Newlines are considered non-black.
 */

int skipgetc(fp)
FILE *fp;
{
register int ch;

	while ((ch = fgetc(fp)) == ' ' || ch == '\t')
		;
	return(ch);
}
