/* EXEC_UTIL -- executive utility routines */
/* (c) 1988 - University of Michigan */

#include <stdio.h>

#include "global.h"
#include "assert.h"
#include "exec.h"

int max_leaf_id = 0;

#ifdef SIZE
long m_tnode = 0;
#endif SIZE

/* INITIALIAZE(prob,crit)
 * This sets up the initial data structure (the root of the search tree) and
 * initializes all other modules.
 */

initialize(prob_fp,crit_fp,debg_fp)
FILE *prob_fp, *crit_fp, *debg_fp;
{
Leaf first;

	/* Initialize the cad interface */
	init_debug(debg_fp);
	init_cad(prob_fp);

	/* Make an initial leaf */
	first = (Leaf)m_alloc(N_TREE_MOD*sizeof(struct Leaf_Data *));
	first[DATA] = (struct Leaf_Data *)m_alloc(sizeof(struct Leaf_Data));
	first[DATA]->id = max_leaf_id++;
#ifdef DISCARD
	first[DATA]->collapsed = FALSE;
	first[DATA]->tree = NULL;
#endif DISCARD

	/* Initialize the submodules */
	crit_init(first,crit_fp);
	/* dump_weights(stdout); */
	acg_init(first);
	if (immediate_consequence(first))
	{
		printf("Problem is not Solvable\n");
		return;
	}

	/* Shut down cad interface */
	dispose_cad();

	/* Get initial ratings and all */
	first[DATA]->rating = crit_rate(first);
	first[DATA]->completion = acg_rate(first);
	first[DATA]->depth = 0;
#ifdef CRIPPLE   
	if (first[DATA]->completion > 0) 
	      first[DATA]->rating = 99999;
#endif CRIPPLE

	/* Initialize list of leaves */
	init_list(first);

	/* Dump first leaf */
	if (is_on(SHOW_CONSTRAINTS)) dump_constraints(first,stdout);
	if (is_on(SHOW_LEAVES) && !is_on(SHOW_EXPAND)) dump_leaf(first,stdout);
			 
}

/* COPY_LEAF(old)
 * Return a pointer to a new leaf which is a copy of the old leaf.  Both the
 * old leaf and the new leaf are assigned new id numbers, though, and both
 * depths are one more than the old leaf's depth.  If DISCARD is on, the
 * tree is updated so that both output nodes are children of the input node.
 */

Leaf copy_leaf(old)
Leaf old;
{
Leaf new;

	new = (Leaf)m_alloc(N_TREE_MOD*sizeof(struct Leaf_Data *));
	new[DATA] = (struct Leaf_Data *)m_alloc(sizeof(struct Leaf_Data));
	new[DATA]->rating = old[DATA]->rating;
	new[DATA]->completion = old[DATA]->completion;
	old[DATA]->id = max_leaf_id++;
	new[DATA]->id = max_leaf_id++;
	new[DATA]->depth = ++old[DATA]->depth;

	acg_copy(new,old);
	crit_copy(new,old);

#ifdef DISCARD
	new[DATA]->tree = old[DATA]->tree;
	new[DATA]->collapsed = old[DATA]->collapsed;
#endif DISCARD

	return(new);
}


/* CLONE_LEAF(old)
 * Return a pointer to a new leaf which is an exact copy of the old leaf.
 * The original leaf is unaltered.
 */

Leaf clone_leaf(old)
Leaf old;
{
Leaf new;

	new = (Leaf)m_alloc(N_TREE_MOD*sizeof(struct Leaf_Data *));
	new[DATA] = (struct Leaf_Data *)m_alloc(sizeof(struct Leaf_Data));
	new[DATA]->rating = old[DATA]->rating;
	new[DATA]->completion = old[DATA]->completion;
	new[DATA]->id = old[DATA]->id;
	new[DATA]->depth = old[DATA]->depth;
#ifdef DISCARD
	new[DATA]->tree = old[DATA]->tree;
	new[DATA]->collapsed = old[DATA]->collapsed;
#endif DISCARD

	acg_copy(new,old);
	crit_copy(new,old);

	return(new);
}

/* DROP_LEAF(leaf)
 * Deallocate the memory for the given leaf.
 */

drop_leaf(leaf)
Leaf leaf;
{
	acg_drop(leaf);
	crit_drop(leaf);
	/* We should really work our way up the tree discarding as much as
	   we can, but at the moment, we are too lazy for that. */
#ifdef DISCARD
	if (leaf[DATA]->tree != NULL)
		free_tnode(leaf[DATA]->tree);
#endif DISCARD
	m_free(leaf[DATA],sizeof(struct Leaf_Data));
	m_free(leaf,N_TREE_MOD*sizeof(struct Leaf_Data *));
}

#ifdef DISCARD
/* COLLAPSE_LEAF(leaf)
 * Remember we had this leaf, but throw out its acg and criteria data
 * structures.
 */

collapse_leaf(leaf)
Leaf leaf;
{
	acg_drop(leaf);
	crit_drop(leaf);
	leaf[DATA]->collapsed = TRUE;
}


/* SAVE_PROP(leaf,prop)
 * This copies a proposal into the storage built into a tree node.
 */

save_prop(leaf,prop)
Leaf leaf;
Prop *prop;
{
register Tnode *tree;
int m;

	if (prop->type == SEQU)
	{
		/* Save a sequencing proposal */
		tree = (Tnode *)m_alloc(m = sizeof(Tnode) -
			(sizeof(TrajProp) - sizeof(SequProp)));
		tree->type = SEQU;
		tree->store.s.p1 = prop->a.sequ->p1;
		tree->store.s.p2 = prop->a.sequ->p2;
	}
	else if (cardinality(&prop->a.traj->t) == 1)
	{
		/* Save a trajectory proposal with only one traj */
		tree = (Tnode *)m_alloc(m = sizeof(Tnode) -
			(sizeof(TrajProp) - sizeof(SequProp)));
		tree->type = 't';
		tree->store.s.p1 = prop->a.traj->p;
		tree->store.s.p2 = first_in(&prop->a.traj->t);
	}
	else
	{
		/* Save a full blown trajectory proposal */
		tree = (Tnode *)m_alloc(m = sizeof(Tnode));
		tree->type = TRAJ;
		tree->store.t.p = prop->a.traj->p;
		setcopy(&tree->store.t.t,&prop->a.traj->t);
	}

	/* Push it on the list */
	tree->papa = leaf[DATA]->tree;
	leaf[DATA]->tree = tree;
#ifdef SIZE
	m_tnode += m;
#endif SIZE
}

free_tnode(tree)
Tnode *tree;
{
int m;

    if (tree->type == TRAJ)
	m_free(tree,m = sizeof(Tnode));
    else
	m_free(tree,m = sizeof(Tnode) - (sizeof(TrajProp) - sizeof(SequProp)));
#ifdef SIZE
    m_tnode -= m;
#endif SIZE
}


/* GET_TREE_PROP(tree)
 * This returns a pointer to a copy of the proposal in the given tree node.
 * The proposal is stored in local static storage, so it is murdered by
 * the next call to this routine.
 */
Prop *get_tree_prop(tree)
Tnode *tree;
{
static Prop p;
static SequProp s;
static TrajProp t;

	if (tree->type == TRAJ)
	{
		p.type = TRAJ;
		p.a.traj = &t;
		t.p = tree->store.t.p;
		setcopy(&t.t,&tree->store.t.t);
	}
	else if (tree->type == SEQU)
	{
		p.type = SEQU;
		p.a.sequ = &s;
		s.p1 = tree->store.s.p1;
		s.p2 = tree->store.s.p2;
	}
	else if (tree->type == 't')
	{
		p.type = TRAJ;
		p.a.traj = &t;
		t.p = tree->store.s.p1;
		settrajset(&t.t,tree->store.s.p2);
	}
	else
		error("Bad Proposal Type %c in Tree\n",tree->type);
	return(&p);
}
#endif DISCARD


#ifdef SIZE
long exec_leaf_size(leaf)
Leaf leaf;
{
	return ((long)(sizeof(struct Leaf_Data) +
		       N_TREE_MOD*sizeof(struct Leaf_Data *)));
}

#ifdef DISCARD
long exec_tree_size()
{
	return(m_tnode);
}
#endif DISCARD
#endif SIZE
