/* LIST -- List management module */
/* (c) 1988 - University of Michigan */

#include <stdio.h>
#include "global.h"
#include "list.h"
#include "exec.h"

#ifdef HEAP

/**************************************************************************/
/*                             H E A P S                                  */
/**************************************************************************/

struct heap_desc {
	Leaf *d;	/* Pointer to array of pointers to leaves */
	int max;	/* Size of array */
	int size;	/* Current number of leaves in heap */
	} heap[N_HEAP];

#define INITMAX 64	/* Size to start heaps at */

/* WORSE(a,b) -- Is leaf a worse than leaf b? */
#define worse(a,b) (a[DATA]->rating > b[DATA]->rating || (a[DATA]->rating == b[DATA]->rating && (a[DATA]->completion > b[DATA]->completion || a[DATA]->completion == b[DATA]->completion && a[DATA]->id < b[DATA]->id)))

/* BELOW(a,b,h) -- Should leaf a be above leaf b in heap h? */
#define below(a,b,h) (worse(a,b)? !h : h)

Leaf rm_root_from_heap();

/* -------------- Level One:  Module Entries ---------------- */

/* INIT_LIST
 * Put the one leaf given on the initial heaps.
 */

void init_list(leaf)
Leaf leaf;
{
	/* Initialize ALL heap */
	heap[ALL].max = INITMAX;
	heap[ALL].size = 1;
	heap[ALL].d = (Leaf *)m_alloc(INITMAX * sizeof(Leaf *));
	heap[ALL].d[0] = leaf;
	leaf[DATA]->l.heap[ALL] = 0;

#ifdef DISCARD
	/* Initialize OPEN heap */
	heap[OPEN].max = INITMAX;
	heap[OPEN].size = 1;
	heap[OPEN].d = (Leaf *)m_alloc(INITMAX * sizeof(Leaf *));
	heap[OPEN].d[0] = leaf;
	leaf[DATA]->l.heap[OPEN] = 0;
#endif DISCARD
}

#ifdef DISCARD
/* RM_WORST_OPEN
 * Return the lowest rated open leaf, and delete it from the list of open
 * leaves.  This does not delete the node from the list of all nodes.
 */
Leaf rm_worst_open()
{
	return(rm_root_from_heap(OPEN));
}
#endif DISCARD

/* RM_BEST
 * Remove the best node from the list of nodes.
 */

Leaf rm_best()
{
register Leaf best;

	best = rm_root_from_heap(ALL);

#ifdef DISCARD
	/* If this is an open leaf, remove it from the open heap too */
	if (best != NULL && best[DATA]->l.heap[OPEN] != -1)
		/* It has to be a terminal leaf in the open heap */
		rm_term_from_heap(OPEN,best);
#endif DISCARD
	return(best);
}


/* GET_BEST
 * Return the best node from the list of nodes.
 */

Leaf get_best()
{
	return(heap[ALL].d[0]);
}

/* ADD_OPEN
 * Add a new open leaf to the data structure
 */

void add_open(leaf)
Leaf leaf;
{
	add_to_heap(ALL,leaf);
#ifdef DISCARD
	add_to_heap(OPEN,leaf);
#endif DISCARD
}


/* CHANGE_BEST
 * This is called to indicate that the rating of the best node has been
 * changed.  It's position in the lists should be changed accordingly.
 */
void change_best()
{
#ifdef DISCARD
Leaf l = heap[ALL].d[0];

	/* Best leaf of all must be a terminal leaf of the open   */
	/* heap so we only have to check above it for heapishness */
	if (l[DATA]->l.heap[OPEN] == -1)
		/* Newly rebuilt leaves may not be on open heap yet */
		add_to_heap(OPEN,l);
	else
		upheap(OPEN,l[DATA]->l.heap[OPEN]);
#endif DISCARD
	downheap(ALL,0);
}

/* -------------- Level Two:  Heap Operations ---------------- */

add_to_heap(h,l)
int h;
Leaf l;
{
register int sz;

	/* If heap is full, double its size */
	if ((sz = heap[h].size) == heap[h].max)
	    heap[h].d = (Leaf *)m_realloc( heap[h].d,
					sz * sizeof(Leaf *),
					(heap[h].max = sz*2)*sizeof(Leaf *));
	heap[h].d[sz] = l;
	l[DATA]->l.heap[h] = sz;
	heap[h].size = ++sz;
	upheap(h,sz-1);
}

Leaf rm_root_from_heap(h)
int h;
{
register Leaf best,tmp;

	if (heap[h].size == 0) return(NULL);

	/* Peel best node off top of heap */
	best = heap[h].d[0];
	best[DATA]->l.heap[h] = -1;

	/* Move last heap element into root position and decrement size */
	tmp = heap[h].d[0] = heap[h].d[--heap[h].size];
	tmp[DATA]->l.heap[h] = 0;

	/* Fix the heap */
	downheap(h,0);

	return(best);
}

#ifdef DISCARD
/* This can be used to throw away any terminal node (node with no children)
 * on the heap.  */
rm_term_from_heap(h,l)
int h;
Leaf l;
{
register Leaf tmp;
register int k;

	if (heap[h].size == 0 || l[DATA]->l.heap[h] == -1 ||
		l[DATA]->l.heap[h] < (heap[h].size>>1))
		error("nonterminal leaf %d\n",l[DATA]->id);

	k = l[DATA]->l.heap[h];
	l[DATA]->l.heap[h] = -1;

	/* Move last heap element into the position of the deleted leaf */
	tmp = heap[h].d[k] = heap[h].d[--heap[h].size];
	tmp[DATA]->l.heap[h] = k;

	/* Fix the heap */
	upheap(h,k);
}
#endif DISCARD

/* -------------- Level Three:  Heap Primitives ---------------- */

upheap(h,k)
register int h,k;
{
register Leaf v;
register int j;

	v = heap[h].d[k];
	while (k > 0 && below(heap[h].d[j = (k-1)>>1],v,h))
	{
		(heap[h].d[k] = heap[h].d[j])[DATA]->l.heap[h] = k;
		k = j;
	}
	heap[h].d[k] = v;
	v[DATA]->l.heap[h] = k;
}

downheap(h,k)
register int h,k;
{
register Leaf v;
register int j;
int m,sz;

	v = heap[h].d[k];
	m = (sz = heap[h].size) >> 1;
	while (k < m)
	{
		j = k + k + 1;
		if (j < sz-1 && below(heap[h].d[j],heap[h].d[j+1],h)) j++;
		if (!below(v,heap[h].d[j],h)) break;
		(heap[h].d[k] = heap[h].d[j])[DATA]->l.heap[h] = k;
		k = j;
	}
	heap[h].d[k] = v;
	v[DATA]->l.heap[h] = k;
}

/* --------------  Debug Junk ---------------- */

dump_tree(fp)
FILE *fp;
{
register int i;

	if (heap[ALL].size == 0)
		fprintf(fp,"Empty Search Tree\n");
	for (i=0; i<heap[ALL].size; i++)
		exec_dump(heap[ALL].d[i],fp);
}


dump_fulltree(fp)
FILE *fp;
{
register int i;

	if (heap[ALL].size == 0)
                fprintf(fp,"Empty Search Tree\n");
	for (i=0; i<heap[ALL].size; i++)
                dump_leaf(heap[ALL].d[i],fp);
}

 
#ifdef DISCARD
dump_open(fp)
FILE *fp; 
{
register int i;
 
	if (heap[OPEN].size == 0)
                fprintf(fp,"\nOpen List is Empty.\n");
        else
        {  
                fprintf(fp,"\nOpen List:\n");
		for (i=0; i<heap[OPEN].size; i++)
			exec_dump(heap[OPEN].d[i],fp);
        }
}
#endif DISCARD

 
#ifdef SIZE
dump_size(fp)
FILE *fp;
{
Leaf l;
int i;
long sz,st;
long crit_stat_size(),acg_stat_size(),cad_size();
#ifdef DISCARD
long exec_tree_size();
#endif DISCARD
 
        sz = 0;
 
	for (i=0; i<heap[ALL].size; i++)
        {
		l = heap[ALL].d[i];
                fprintf(fp,"LEAF %d SIZE:\n", l[DATA]->id);
                sz += dump_leaf_size(fp,l);
        }

#ifdef DISCARD
        fprintf(fp,"LEAF ORIG SIZE:\n");
        sz += st = dump_leaf_size(fp,orig);
        fprintf(fp,"leaf total size = %ld\n", sz);
        sz += st = (heap[ALL].max + heap[OPEN].max)*sizeof(Leaf *);
        fprintf(fp,"HEAP size = %ld\n",st);
        sz += st = exec_tree_size();
        fprintf(fp,"EXEC tree size = %ld\n",st);
#else
        fprintf(fp,"leaf total size = %ld\n", sz);
        sz += st = heap[ALL].max*sizeof(Leaf *);
        fprintf(fp,"HEAP size = %ld\n",st);
#endif DISCARD
        sz += st = cad_size();
        fprintf(fp,"CAD static size = %ld\n",st);
        sz += st = acg_stat_size();
        fprintf(fp,"ACG static size = %ld\n",st);
        sz += crit_stat_size(fp);
        fprintf(fp,"TOTAL SIZE = %ld\n\n",sz);
}
#endif SIZE

#else

/**************************************************************************/
/*                    S O R T E D   L I S T                               */
/**************************************************************************/

Leaf best_leaf;		/* best leaf (other leafs hang off in linked list) */
#ifdef DISCARD
Leaf worstopen;         /* worst uncollapsed leaf */
#endif DISCARD

/* INIT_LIST
 * Put the one leaf given on the initial list.
 */

void init_list(leaf)
Leaf leaf;
{
	best_leaf = leaf;
#ifdef DISCARD
	leaf[DATA]->l.prevopen = leaf;
	leaf[DATA]->l.nextopen = leaf;
	worstopen = leaf;
#endif DISCARD
}


#ifdef DISCARD
/* RM_WORST_OPEN
 * Return the lowest rated open leaf, and delete it from the list of open
 * leaves.  Probably this should call rm_from_open
 */
Leaf rm_worst_open()
{
Leaf tmp;

	if (worstopen == NULL) return (NULL);

	tmp = worstopen;
	if (tmp[DATA]->l.nextopen == tmp)
	{
		worstopen = NULL;
	}
	else
	{
		worstopen = tmp[DATA]->l.nextopen[DATA]->l.prevopen =
			tmp[DATA]->l.prevopen;
		worstopen[DATA]->l.nextopen = tmp[DATA]->l.nextopen;
	}

	tmp[DATA]->l.prevopen = NULL;		/* Mark it not on open list */
	return(tmp);
}
#endif DISCARD

/* RM_BEST
 * Remove the best node from the list of nodes.
 */

Leaf rm_best()
{
Leaf tmp = best_leaf;

	if (best_leaf != NULL)
		best_leaf = best_leaf[DATA]->l.next;
#ifdef DISCARD
	rm_from_open(tmp);
#endif DISCARD
	return(tmp);
}


/* GET_BEST
 * Return the best node from the list of nodes.
 */

Leaf get_best()
{
	return(best_leaf);
}

/* ADD_OPEN
 * Add a new open leaf to the data structure
 */

void add_open(leaf)
Leaf leaf;
{
Leaf preopen,pred;

	if (pred = find_pred(leaf,best_leaf,&preopen))
	{
		leaf[DATA]->l.next = pred[DATA]->l.next;
		pred[DATA]->l.next = leaf;
	}
	else
	{  
		leaf[DATA]->l.next = best_leaf;
		best_leaf = leaf;
	}
#ifdef DISCARD
	insertopen(leaf,preopen);
#endif DISCARD
}


/* CHANGE_BEST
 * This is called to indicate that the rating of the best node has been
 * changed.  It's position in the lists should be changed accordingly.
 */
void change_best()
{
Leaf tmp, preopen,pred;

	tmp = best_leaf;
#ifdef DISCARD
	rm_from_open(tmp);
#endif DISCARD

	/* Insert in in the list */
	if (pred = find_pred(best_leaf,best_leaf[DATA]->l.next,&preopen))
	{
		best_leaf = best_leaf[DATA]->l.next;
		tmp[DATA]->l.next = pred[DATA]->l.next;
		pred[DATA]->l.next = tmp;
	}
#ifdef DISCARD  
	insertopen(tmp,preopen);
#endif DISCARD
}

#ifdef DISCARD
/* RM_FROM_OPEN
 * Take the given node off of the open leaf list.  Internal use only.
 */

rm_from_open(leaf)
Leaf leaf;
{
	/* If leaf is not on open list return */
	if (leaf[DATA]->l.prevopen == NULL) return;
        
	/* If this is only open leaf, empty open list */
	if (leaf[DATA]->l.prevopen == leaf)
	{
		worstopen = NULL;
		return;
	}

	/* Delete leaf from open list */
	leaf[DATA]->l.prevopen[DATA]->l.nextopen = leaf[DATA]->l.nextopen;
	leaf[DATA]->l.nextopen[DATA]->l.prevopen = leaf[DATA]->l.prevopen;
	leaf[DATA]->l.prevopen = NULL;	/* Mark that it is not on open */

	/* If this was worst open leaf, adjust worstopen pointer */
        if (worstopen == leaf)
                worstopen = leaf[DATA]->l.prevopen;
}


/* INSERTOPEN
 * Insert leaf after leaf preopen in the open list.  If preopen is NULL,
 * put it in front.  Internal use only.
 */
insertopen(leaf,preopen)
Leaf leaf,preopen;
{
        if (worstopen == NULL)
        {
                leaf[DATA]->l.prevopen = leaf;
                leaf[DATA]->l.nextopen = leaf;
                worstopen = leaf;
        }
        else
        {   
                if (preopen == worstopen)
                        worstopen = leaf;
                else if (preopen == NULL)
                        preopen = worstopen;
                leaf[DATA]->l.prevopen = preopen;
                leaf[DATA]->l.nextopen = preopen[DATA]->l.nextopen;
                preopen[DATA]->l.nextopen = leaf;
                leaf[DATA]->l.nextopen[DATA]->l.prevopen = leaf;
        }
}
#endif DISCARD

 
/* FINDPRED
 * Search the list of leaves headed by "head" and find the one after which
 * "leaf" should be inserted.  Return NULL if it should be the the first on
 * the list.  If DISCARD is defined, the previous open leaf is returned in
 * prevopen.  NULL is returned there if this is the first open leaf.
 * Internal use only.
 */

Leaf find_pred(leaf,head,prevopen)
Leaf leaf,head;
Leaf *prevopen;
{
Leaf prev=NULL;
int rating=leaf[DATA]->rating;
int comple=leaf[DATA]->completion;
 
        *prevopen = NULL;
        while (head != NULL && (rating > head[DATA]->rating ||
            (rating == head[DATA]->rating && comple > head[DATA]->completion)))
        {
#ifdef DISCARD
                if (!head[DATA]->collapsed)
                        *prevopen = head;
#endif DISCARD
                prev = head;
                head = head[DATA]->l.next;
        }
        return(prev);
}

dump_tree(fp)
FILE *fp;
{
Leaf p;

	if (best_leaf==NULL)
		fprintf(fp,"Empty Search Tree\n");
	for (p=best_leaf; p!=NULL; p=p[DATA]->l.next)
		exec_dump(p,fp);
}


dump_fulltree(fp)
FILE *fp;
{
Leaf l;

        if (best_leaf==NULL)
                fprintf(fp,"Empty Search Tree\n");
        for (l=best_leaf; l!=NULL; l=l[DATA]->l.next)
                dump_leaf(l,fp);
}

 
#ifdef DISCARD
dump_open(worst,fp)
Leaf worst;
FILE *fp; 
{
Leaf tmp;
 
        if (worst == NULL)
                fprintf(fp,"\nOpen List is Empty.\n");
        else
        {  
                fprintf(fp,"\nOpen List:\n");
                exec_dump(worst,fp);  
                for (tmp=worst[DATA]->l.prevopen;
                     tmp != worst;    
                     tmp=tmp[DATA]->l.prevopen)
                {
                        exec_dump(tmp,fp);
                        if (tmp[DATA]->l.nextopen[DATA]->l.prevopen != tmp)
                                fprintf(fp,"(nextopen pointer inconsistant)\n");                }
        }
}
#endif DISCARD

 
#ifdef SIZE
dump_size(fp)
FILE *fp;
{
Leaf l;
long sz,st;
long crit_stat_size(),acg_stat_size(),cad_size();
#ifdef DISCARD
long exec_tree_size();
#endif DISCARD
 
        sz = 0;
 
        for (l=best_leaf; l!=NULL; l=l[DATA]->l.next)
        {
                fprintf(fp,"LEAF %d SIZE:\n", l[DATA]->id);
                sz += dump_leaf_size(fp,l);
        }

#ifdef DISCARD
        fprintf(fp,"LEAF ORIG SIZE:\n");
        sz += st = dump_leaf_size(fp,orig);
        fprintf(fp,"leaf total size = %ld\n", sz);
        sz += st = exec_tree_size();
        fprintf(fp,"EXEC tree size = %ld\n",st);
#else
        fprintf(fp,"leaf total size = %ld\n", sz);
#endif DISCARD
        sz += st = cad_size();
        fprintf(fp,"CAD static size = %ld\n",st);
        sz += st = acg_stat_size();
        fprintf(fp,"ACG static size = %ld\n",st);
        sz += crit_stat_size(fp);
        fprintf(fp,"TOTAL SIZE = %ld\n\n",sz);
}
#endif SIZE

#endif HEAP
