/* MANI.C -- Manipulability criterion */
/* (c) 1988 - University of Michigan */

#include <stdio.h>
#include "config.h"
#include "set.h"
#include "global.h"
#include "assert.h"

#define DATA MANI_MOD

/* ----------------------------------------------------------------------- */
/*           D A T A   S T R U C T U R E   D E C L A R A T I O N S         */
/* ----------------------------------------------------------------------- */

/* DIFF type, and DIFFSETs
 * This is a type for difficulty class numbers.  A class with a lower number
 * should also have a lower difficulty rating.  DiffSets hold sets of
 * difficulty classes.
 */

typedef int Diff;

#define MAXDIFFS MAXTRAJS
defset(DiffSet,MAXDIFFS);
#define mkdiffset() makeset(n_difficulties,EOS)
#define emptydiffset(s) initset(s,n_difficulties,EOS)
#define setdiffset(s,e) initset(s,n_difficulties,e,EOS)

/* Static structures */

Diff diff_array[MAXTRAJS];	/* Difficulty classes of trajectories */
int rate_array[MAXDIFFS];	/* Difficulty ratings of difficulty classes */
int mani_array[MAXPARTS];	/* Manipulability ratings of parts */

int n_difficulties;

/* Local Leaf Data */

struct pdc {
	DiffSet ds;		/* Set of difficulty classes part could be in */
	int nds;		/* No. of difficulty classes part could be in */
	};

struct Leaf_Data {
#ifdef MERGE1
	int link_cnt;	 /* Number of leaves using this data */
#endif MERGE1
	struct pdc *dc;  /* Array of pointers to eds lists indexed by part */
			 /* NULL after retirement */
	int rating;	 /* Rating.  -1 if not computed yet */
	};

/* ----------------------------------------------------------------------- */
/*        D A T A   S T R U C T U R E   A C C E S S   R O U T I N E S      */
/* ----------------------------------------------------------------------- */

#define diff_number() n_difficulties

/* DIFF_TRAJ(traj), MANI_PART(part), 
 * Get difficulty class or rating of a trajectory or manipulability of a part.
 */

#define diff_traj(t) diff_array[t]
#define rate_diff(d) rate_array[d]
#define rate_traj(t) rate_diff(diff_traj(t))
#define mani_part(p) mani_array[p]

/* INIT_DIFFS()
 * Set up the static data structures
 */
init_diffs()
{
register int dr,tmp_dr;
register Diff dc;
Traj t;
Part p;

    /* Build up the class->difficulty mapping, using a bubble sort */
    n_difficulties = 0;
    for (t=0; t<traj_number(); t++)
    {
	dr = diff_array[t] = get_diff(t);
	for (dc=0; dc < n_difficulties; dc++)
	    if (rate_array[dc] == dr)
		goto next;
	    else if (rate_array[dc] > dr)
	    {
		tmp_dr = dr;
		dr = rate_array[dc];
		rate_array[dc] = tmp_dr;
	    }
	rate_array[n_difficulties++] = dr;
next:;
    }

    /* Fix up the traj->class mapping */
    for (t=0; t<traj_number(); t++)
    {
	dr = diff_array[t];
	for (dc=0; dc < n_difficulties; dc++)
	    if (rate_array[dc] == dr)
		diff_array[t] = dc;
    }

    /* Get the part manipulability */
    for (p=0; p<part_number(); p++)
	mani_array[p] = get_mani(p);
}		

/* DIFFS_TRAJS(ds,ts)
 * Convert a trajectory set to a difficulty set, return the cardinality of
 * the diffset.
 */

int diffs_trajs(ds,ts)
DiffSet *ds;
TrajSet *ts;
{
register Traj t;
register Diff d;
register int card = 0;

	emptydiffset(ds);
	foreach(t,ts)
		if (!isin(ds,d = diff_traj(t)))
		{
			append1(ds,d);
			card++;
		}
	return(card);
}

/* BEST_DIFF(ds), WORST_DIFF(ds)
 * Return the best or worst difficulty in the given set.
 */

#define best_diff(d) first_in(d)
#define worst_diff(d) last_in(d)


/* GET_DIFFS(leafdata,part,set,card)
 * Get a pointer to the set of possible difficulty classes for a part and a
 * pointer to the cardinality of that set.  To change the set, just change
 * whatever these point to and call reset_diffs().  The contents of these
 * memory locations should not be changed unless we intend to call
 * reset_diffs() on it.
 */

#define get_diffs(d,p,z,m) (z = &d->dc[p].ds, m = &d->dc[p].nds)

/* RESET_DIFFS()
 * Changes whatever leaf,part was last get_diffs()ed to whatever is now in the
 * the set whose pointer was returned.
 */

#define reset_diffs() /* nop */

/* GET_MANI_LINKS(leaf), SET_MANI_LINKS(leaf,links)
 * Get, set or decrement the link_cnt.  New value is always returned.
 */

#ifdef MERGE1
#define get_mani_links(l) (l[DATA]->link_cnt)
#define set_mani_links(l,c) (l[DATA]->link_cnt = c)
#define dec_mani_links(l) (--(l[DATA]->link_cnt))
#define inc_mani_links(l) (++(l[DATA]->link_cnt))
#else
#define get_mani_links(l) 1
#define set_mani_links(l,c) 1
#define dec_mani_links(l) 0
#define inc_mani_links(l) 1
#endif MERGE1


/* IS_RETIRED(leaf)
 * Check if this leaf is retired.
 */

#ifdef RETIRE
#define is_retired(l) (l[DATA]->dc == NULL)
#else
#define is_retired(l) FALSE
#endif RETIRE


/* MANI_ALLOC(leaf)
 * Allocate memory for a leaf.  Don't initialize it.  Both the leaf_data and
 * the array are allocated.
 *
 * MANI_RETIRE(leaf)
 * Deallocate the array for the leaf and mark it as retired.
 *
 * MANI_FREE(leaf)
 * Deallocate the leaf_data for the leaf.
 */

mani_alloc(leaf)
Leaf leaf;
{
   leaf[DATA] = (struct Leaf_Data *)m_alloc(sizeof(struct Leaf_Data));
   leaf[DATA]->dc = (struct pdc *)m_alloc(sizeof(struct pdc) * part_number());
}

mani_retire(leaf)
Leaf leaf;
{
    if (is_on(SHOW_ENTRIES)) printf("mani_retire\n");
    if (is_retired(leaf)) return;
    m_free(leaf[DATA]->dc,sizeof(struct pdc) * part_number());
    leaf[DATA]->dc = NULL;
}

#define mani_free(l) m_free(l[DATA],sizeof(struct Leaf_Data))


#define UNKNOWN -1

/* GET_MANI_RATING(leaf), SET_MANI_RATING(leaf,rating)
 * Get or set the old rating.  UNKNOWN is a legal value.  Both return the
 * new value.
 */

#ifdef RETIRE
#define get_mani_rating(l) (l[DATA]->rating)
#define set_mani_rating(l,r) (l[DATA]->rating = r)
#else
#define get_mani_rating(l) 0
#define set_mani_rating(l,r) r
#endif

/* ----------------------------------------------------------------------- */
/*                   H I G H   L E V E L   R O U T I N E S                 */
/* ----------------------------------------------------------------------- */

/* MANI_INIT(leaf)
 * Initialize the data structure for the root leaf, allocating memory and
 * loading it up with data from CAD.
 */

mani_init(leaf)
Leaf leaf;
{
int parts = part_number();
int trajs = traj_number();
boolean should_retire = TRUE;
int dif;
DiffSet *ds;
int *cp;
Part p;
Traj t;

    if (is_on(SHOW_ENTRIES)) printf("mani_init\n");

    init_diffs();
    mani_alloc(leaf);
    set_mani_links(leaf,1);

#ifdef RETIRE
    if (diff_number() > 1)
#endif RETIRE
	for (p=0; p<parts; p++)
	{
	    get_diffs(leaf[DATA],p,ds,cp);
	    if ((*cp = diffs_trajs(ds,proposed_traj(p))) > 1)
		    should_retire = FALSE;
	    reset_diffs();
	}

#ifdef RETIRE
    /* Check if we should retire it right away */
    if (should_retire)
    {
	set_mani_rating(leaf,mani_rate(leaf));
	mani_retire(leaf);
    }
#endif RETIRE

    set_mani_rating(leaf,UNKNOWN);
}


/* MANI_COPY(new,old)
 * Copy the given leaf into a new one.  This is done by making both point to
 * the same leaf if MERGE1 is defined.  Otherwise we make a physical copy now.
 *
 * MANI_SPLIT(leaf)
 * This takes a leaf with a link count greater than one and divides it into a
 * separate leaf with it's own memory.  Used if MERGE1 is defined only.
 */

#ifdef MERGE1
mani_copy(new,old)
Leaf new,old;
{
	new[DATA] = old[DATA];
	inc_mani_links(new);
}

mani_split(new)
Leaf new;
{
Leaf old = new;
#else
#define mani_split(l) /* nop */
mani_copy(new,old)
Leaf new,old;
{
#endif MERGE1
DiffSet *ods, *nds;
int *ocp, *ncp;
int parts = part_number();
struct Leaf_Data *nl;
struct Leaf_Data *ol=old[DATA];
Part p;

    if (is_on(SHOW_ENTRIES)) printf("mani_split\n");
    if (is_retired(old)) error("Attempt to mani_split a retired leaf\n");

    /* Separate the two leaves */
    mani_alloc(new);
    nl = new[DATA];
#ifdef MERGE1
    ol->link_cnt--;
    nl->link_cnt = 1;
#endif MERGE1

    /* Copy the data */

    nl->rating = ol->rating;

    for (p = 0; p < parts; p++)
    {
	get_diffs(ol,p,ods,ocp);
	get_diffs(nl,p,nds,ncp);
	setcopy(nds,ods);
	*ncp = *ocp;
	reset_diffs();
    }
} /* } */


/* MANI_DROP(leaf)
 * Release all memory used by our data structure in the given leaf, if this is
 * the only link to the leaf.  Otherwise just decrement the link count.
 */

mani_drop(leaf)
Leaf leaf;
{
    if (is_on(SHOW_ENTRIES)) printf("mani_drop\n");

    if (dec_mani_links(leaf) == 0)
    {
	mani_retire(leaf);
	mani_free(leaf);
    }
    leaf[DATA] = NULL;	/* make sure we get an error if we use this again */
}


/* MANI_DUMP(leaf,fp)
 * print the contents of the our leaf data on the given device.
 */

mani_dump(leaf,fp)
Leaf leaf;
FILE *fp;
{
Part p;
DiffSet *ds;
Diff d;
int *cp;
int parts = part_number();

    if (is_retired(leaf))
	fprintf(fp,"Manipulability Retired. (lc=%d) rating = %d\n",
			get_mani_links(leaf), get_mani_rating(leaf));
    else
    {
	fprintf(fp,"Manipulability. (lc=%d) rating = %d\n",
			get_mani_links(leaf), get_mani_rating(leaf));
	for (p = 0; p < parts; p++)
	{
	    get_diffs(leaf[DATA],p,ds,cp);
	    fprintf(fp,"  Part %s manip = %d diff(%d) =",
		part_name(p),mani_part(p),*cp);
	    foreach (d,ds)
		fprintf(fp," %d",rate_diff(d));
	    fprintf(fp,"\n");
	}
    }
}


/* MANI_TRAJ(leaf,ta)
 * MANI_SEQU(leaf,sa)
 * Update the data structure to account for the given assertions.
 */

mani_traj(leaf,ta)
Leaf leaf;
register TrajAsrt *ta;
{
DiffSet *ds;
DiffSet tds;
int *cp;
int tc;

    if (is_on(SHOW_ENTRIES)) printf("mani_traj\n");

    /* If retired, ignore updates */
    if (is_retired(leaf)) return;

    /* Get the old and new difficulties, and see if anything has changed */
    get_diffs(leaf[DATA],ta->p,ds,cp);
    if ((tc = diffs_trajs(&tds,&ta->t)) == *cp) return;

    /* Something has changed, so uniquify the leaf and toss its old rating */
    if (get_mani_links(leaf) > 1)
    {
	mani_split(leaf);
	get_diffs(leaf[DATA],ta->p,ds,cp);
    }
    set_mani_rating(leaf,UNKNOWN);

    /* Save the new difficulties set */
    setcopy(ds,&tds);
    *cp = tc;
    reset_diffs();
}


mani_sequ(leaf,sa)
Leaf leaf;
SequAsrt *sa;
{
    /* Do Nothing */
}


/* MANI_RATE(leaf)
 * Make an optimistic estimate of the rating of this plan.
 */

int mani_rate(leaf)
Leaf leaf;
{
int parts = part_number();
register Part p;
DiffSet *ds;
int *cp;
boolean should_retire = TRUE;
int sum = 0;

    if (is_on(SHOW_ENTRIES)) printf("mani_rate\n");

    /* If it is already rated, return that */
    if (get_mani_rating(leaf) != UNKNOWN)
	return(get_mani_rating(leaf));

    /* Add up the best ratings of each part */
    for (p = 0; p < parts; p++)
    {
	get_diffs(leaf[DATA],p,ds,cp);
	sum += rate_diff(best_diff(ds))*mani_part(p);
	if (*cp != 1) should_retire=FALSE;
    }

#ifdef RETIRE
    if (should_retire) mani_retire(leaf);
#endif RETIRE

    return(set_mani_rating(leaf,sum));
}

/* MANI_PROP(leaf,prop)
 * Set prop to a proposal which would be bad for mani.  Return an estimate of
 * how much this proposal would hurt the rating things compared to it's
 * negation. If we find no proposals that would make a difference, we return -1.
 */

int mani_prop(leaf,prop)
Leaf leaf;
Prop *prop;
{
Part p;
Traj t;
int parts = part_number();
DiffSet *ds;
int *cp;
Part best_part;
register int this_score;
int best_score = 0;
static TrajProp tp;
Diff best_d,d;

    if (is_on(SHOW_ENTRIES)) printf("mani_prop\n");
    if (is_retired(leaf)) return(-1); /* If retired, propose nothing */

    for (p = 0; p < parts; p++)
    {
	get_diffs(leaf[DATA],p,ds,cp);

	/* If list has only one difficulty, skip it */
	if (*cp == 1) continue;

	/* Compare the best to the worst difficulty */
	this_score = (rate_diff(d=worst_diff(ds)) - rate_diff(best_diff(ds))) *
		mani_part(p);
	
	if (this_score > best_score)
	{
		best_part = p;
		best_d = d;
		best_score = this_score;
	}
    }

    if (best_score == 0) return(-1);

    prop->type = TRAJ;
    prop->a.traj = &tp;
    tp.p = best_part;
    setcopy(&tp.t,possible_traj(leaf,tp.p));
    foreach(t,&tp.t)
	if (diff_traj(t) == best_d)
		remove1(&tp.t,t);

    return(best_score);
}

#ifdef TEST_PROPOSALS

/* MANI_TEST(leaf,prop)
 * Rate the quality of the given proposal.  Returns an estimate of how much
 * difference it would make.  The rating is positive if the proposal is worse
 * (higher) than it's negation.  It is negative if it is better.
 */

int mani_test(leaf,prop)
Leaf leaf;
Prop *prop;
{
int *cp;
DiffSet *ds;
Part p;
TrajSet *ts1,*ts2,ts;
Traj t;
Diff best_d1,best_d2,d;

	if (is_on(SHOW_ENTRIES)) printf("mani_test\n");

	/* If retired, nothing matters */
	if (is_retired(leaf)) return(0);

	/* Only trajectory proposals matter */
	if (prop->type != TRAJ) return(0);

	p = prop->a.traj->p;

	get_diffs(leaf[DATA],p,ds,cp);

	/* If list has only one difficulty, Proposal doesnt matter */
	if (*cp == 1) return(0);

	/* Get the trajectory set and it's negation */
	ts1 = &prop->a.traj->t;
	setdifference(ts2 = &ts,possible_traj(leaf,p),ts1);

	/* Find lowest difficulty in first set */
	best_d1 = MAXDIFFS+1;
	foreach(t,ts1)
	{
		d = diff_traj(t);
		if (d < best_d1) best_d1 = d;
	}

	/* Find lowest difficulty in second set */
	best_d2 = MAXDIFFS+1;
	foreach(t,ts2)
	{
		d = diff_traj(t);
		if (d < best_d2) best_d2 = d;
	}

	/* Return the change (one of these is the old difficulty) */
	return((rate_diff(best_d1) - rate_diff(best_d2))*mani_part(p));
}
#else
int mani_test() {}
#endif TEST_PROPOSALS

#ifdef SIZE
mani_leaf_size(leaf,size,links)
Leaf leaf;
long *size;
int *links;
{
	*links = get_mani_links(leaf);
	*size = sizeof(struct Leaf_Data);
	if (!is_retired(leaf))
		*size += (sizeof(struct pdc) * part_number());
}

long mani_stat_size()
{
	return((long)0);
}
#endif SIZE
