/* MONO.C -- Monodirectionality criterion */
/* (c) 1988 - University of Michigan */

#include <stdio.h>
#include "config.h"
#include "set.h"
#include "global.h"
#include "assert.h"

#define DATA MONO_MOD

/* Local set type */

#define MAXDIRES MAXTRAJS
defset(DireSet,MAXDIRES);

#define mkdireset() ((DireSet *) makeset(dire_number(),EOS))
#define emptydireset(s) initset(s,dire_number(),EOS)
#define setdireset(s,e) initset(s,dire_number(),e,EOS)

/* Global data structure */

Dire *traj_dir;

/* Leaf data structures */

struct pdirs {
	int ndir;
	DireSet d;
	};

struct Leaf_Data {
#ifdef MERGE1
	int link_cnt;		/* How many leaves are using this */
#endif MERGE1
	struct pdirs *dirs;	/* Array of structures indexed by part no. */
				/* NULL if retired */
	int decided;		/* Number of parts with only one dir */
	DireSet *used;		/* Directions used by some part */
	int nused;		/* Number of directions in used */
	/* following set up by mono_rate for mono_prop and future mono_rates */
	int uncovered;		/* Parts not covered by used dirs */
				/* -1 if not computed yet */
	DireSet *others;	/* Dirs proposed for all uncovered parts */
	};

/* MONO_INIT(leaf)
 * Initialize the data structure for the root leaf, allocating memory and
 * loading it up with data from CAD.
 */

mono_init(leaf)
Leaf leaf;
{
struct pdirs *pd;
int parts = part_number();
int dirs = dire_number();
int trajs = traj_number();
TrajSet *ts, *get_dir_traj();
Dire d;
Traj t;
Part p;
    if (is_on(SHOW_ENTRIES)) printf("mono_init\n");
    /* Allocate and initialize memory for global data structure */
    traj_dir = (Dire *)m_alloc(sizeof(Dire) * trajs);
    for (d=0; d<dirs; d++)
    {
	ts = get_dir_traj(d);
	foreach (t,ts)
		traj_dir[t] = d;
    }

    /* Allocate memory for root leaf */
    leaf[DATA] = (struct Leaf_Data *)m_alloc(sizeof(struct Leaf_Data));
    pd = leaf[DATA]->dirs =
	(struct pdirs *)m_alloc(sizeof(struct pdirs) * parts);
    leaf[DATA]->used = mkdireset();
    leaf[DATA]->others = mkdireset();

    /* Initialize root leaf */
#ifdef MERGE1
    leaf[DATA]->link_cnt = 1;
#endif MERGE1
    leaf[DATA]->decided = 0;
    leaf[DATA]->uncovered = -1;	/* Flag indicates we haven't computed this */
    for (p=0; p<parts; p++)
    {
	ts = proposed_traj(p);
	traj_to_dir(&pd[p].d,ts);
	if ((pd[p].ndir = cardinality(&pd[p].d)) == 1)
	{
		append(leaf[DATA]->used,&pd[p].d);
		leaf[DATA]->decided++;
	}
    }
    leaf[DATA]->nused = cardinality(leaf[DATA]->used);
#ifdef RETIRE
    /* Instantly retire if all parts are decided or all directions are used */
    if (leaf[DATA]->decided == parts || leaf[DATA]->nused == dirs)
	    mono_retire(leaf);
#endif RETIRE
}

traj_to_dir(ds,ts)
register DireSet *ds;
register TrajSet *ts;
{
register Traj t;

    emptydireset(ds);
    foreach(t,ts)
	append1(ds,traj_dir[t]);
}

dir_to_traj(ts,ds)
register TrajSet *ts;
register DireSet *ds;
{
int trajs = traj_number();
register Traj t;

    emptytrajset(ts);
    for (t=0; t<trajs; t++)
	if (isin(ds,traj_dir[t]))
		append1(ts,t);
}


/* MONO_RATE(leaf)
 * Make an optimistic estimate of the rating of this plan.
 */

int mono_rate(leaf)
Leaf leaf;
{
register Part p;
struct pdirs *pd = leaf[DATA]->dirs;
int parts = part_number();
register int ndirs, decided;

    if (is_on(SHOW_ENTRIES)) printf("mono_rate\n");
    /* If all parts are decided, just return the same old rating */
#ifdef RETIRE
    if (pd == NULL)
#else
    if (leaf[DATA]->decided == parts)
#endif RETIRE
	return(leaf[DATA]->nused*parts);

    /* Find all parts not covered by necessary directions */
    if (leaf[DATA]->uncovered < 0)	/* Only if we don't already know it */
    {
	emptydireset(leaf[DATA]->others);
	negate(leaf[DATA]->others);
	leaf[DATA]->uncovered = 0;
	for (p=0; p < parts; p++)
	    if (pd[p].ndir > 1 && isdisjoint(leaf[DATA]->used,&pd[p].d))
	    {
		leaf[DATA]->uncovered++;
		intersect(leaf[DATA]->others,&pd[p].d);
	    }
    }

    /* If all parts are covered, just return the same old rating */
    if (leaf[DATA]->uncovered == 0)
	return(leaf[DATA]->nused*parts);

    if (isempty(leaf[DATA]->others))
	return((leaf[DATA]->nused+2)*parts);
    else
	return((leaf[DATA]->nused+1)*parts);
}

/* MONO_PROP(leaf,prop)
 * Set prop to a proposal which would be bad for mono.  Return an estimate
 * of how much it would hurt the rating things compared to its negation. 
 * If we find no proposals that would make a difference, we return -1.
 */

int mono_prop(leaf,prop)
Leaf leaf;
Prop *prop;
{
register Part p;
int parts = part_number();
struct pdirs *pd = leaf[DATA]->dirs;
Traj t;
TrajSet ts, *tsp;
DireSet ds;
static TrajProp tp;
setsize dircnt[MAXTRAJS];
int maxdircnt;
boolean all_equal;
int dirs;
Dire d,bestd;

    if (is_on(SHOW_ENTRIES)) printf("mono_prop\n");

    /* CASE 0: We are retired */
#ifdef RETIRE
    if (pd == NULL) return(-1);
#else
    if (leaf[DATA]->decided == parts) return(-1);
#endif RETIRE

    prop->type = TRAJ;
    prop->a.traj = &tp;

    /* CASE 1: No uncovered parts */
    if (leaf[DATA]->uncovered == 0)
    {
	/* Search for a part with a possible direction not in covering set */
	for (p = 0; p < parts; p++)
	    if (pd[p].ndir > 1 && !issubset(&pd[p].d,leaf[DATA]->used))
	    {
		/* Propose we use a direction not in the covering set */
		tp.p = p;
		dir_to_traj(&ts,leaf[DATA]->used);
		setdifference(&tp.t,possible_traj(leaf,p),&ts);
		return(parts);
	    }
	 return(-1);
    }

    /* CASE 2: There is another directions that covers all uncovered parts */
    if (!isempty(leaf[DATA]->others))
    {
	/* We would like to use only trajectories in the set ts */
	setunion(&ds,leaf[DATA]->used,leaf[DATA]->others);
	dir_to_traj(&ts,&ds);

	/* Look for parts with trajectories not in ts */
	for (p = 0; p < parts; p++)
	    if (pd[p].ndir > 1 && !issubset(&pd[p].d,&ds))
	    {
		/* Propose we use a trajectory not in ts */
		tp.p = p;
		setdifference(&tp.t,possible_traj(leaf,p),&ts);
		return(parts);
	    }
	 return(-1);
    }

    /* CASE 3: Don't know what the minimal cover might be */
    /* Propose weakly that the most common direction be used */
    dirs = dire_number();
    for (d=0; d<dirs; d++)
	dircnt[d]=0;

    /* Count uncovered parts in which each direction can be used */
    for (p=0; p<parts; p++)
	if (pd[p].ndir > 1 && isdisjoint(leaf[DATA]->used,&pd[p].d))
	    foreach(d,&pd[p].d)
		    dircnt[d]++;

    /* Find the direction most used */
    maxdircnt = 0;
    all_equal = TRUE;
    for (d=0; d<dirs; d++)
	if (dircnt[d] > maxdircnt)
	{
	    all_equal = (maxdircnt == 0);
	    bestd = d;
	    maxdircnt = dircnt[d];
	}
	else if (dircnt[d] != 0 && dircnt[d] < maxdircnt)
	    all_equal = FALSE;

    if (all_equal) return(-1);

    /* Find a part which can use that direction or another */
    for (p=0; p<parts; p++)
	if (pd[p].ndir > 1 && isin(&pd[p].d,bestd))
	{
	    /* Propose using trajectories with other directions */
	    tp.p = p;
	    emptytrajset(&tp.t);
	    tsp = possible_traj(leaf,p);
	    foreach (t,tsp)
		if (traj_dir[t] != bestd)
			append1(&tp.t,t);
	    return(0);	/* zero rating, since we don't trust this much */
	}
}


/* MONO_TRAJ(leaf,ta)
 * MONO_SEQU(leaf,sa)
 * Update the data structure to account for the given assertions.
 */

mono_traj(leaf,ta)
Leaf leaf;
register TrajAsrt *ta;
{
int card;
struct pdirs *pd;
DireSet ds;

    if (is_on(SHOW_ENTRIES)) printf("mono_traj\n");

#ifdef RETIRE
    /* If already retired, ignore updates */
    if (leaf[DATA]->dirs == NULL) return;
#endif RETIRE

    pd = &leaf[DATA]->dirs[ta->p];

    /* If only one direction, can't be any further change */
    if (pd->ndir == 1) return;

    /* If number of directions is the same, no change */
    traj_to_dir(&ds,&ta->t);
    if ((card = cardinality(&ds)) == pd->ndir) return;

#ifdef MERGE1
    /* If we are going to change data structure make sure it isn't shared */
    if (leaf[DATA]->link_cnt > 1)
    {
	mono_split(leaf);
	pd = &leaf[DATA]->dirs[ta->p];
    }
#endif MERGE1

    /* We are going to change this leaf, so toss out old rating info */
    leaf[DATA]->uncovered = -1;

    /* Do update */
    setcopy(&pd->d,&ds);
    if ((pd->ndir = card) == 1)
    {
	append(leaf[DATA]->used,&pd->d);
	leaf[DATA]->nused = cardinality(leaf[DATA]->used);
	leaf[DATA]->decided++;
#ifdef RETIRE
	/* See if we should retire yet */
	if (leaf[DATA]->decided == part_number() ||
	    leaf[DATA]->nused == dire_number())
	    mono_retire(leaf);
#else
#endif RETIRE
    }
}

mono_sequ(leaf,sa)
Leaf leaf;
SequAsrt *sa;
{
    /* Do Nothing */
}


/* MONO_DUMP(leaf,fp)
 * print the contents of the our leaf data on the given device.
 */

mono_dump(leaf,fp)
Leaf leaf;
FILE *fp;
{
Part p;
struct pdirs *pd = leaf[DATA]->dirs;
int parts = part_number();


#ifdef RETIRE
    if (pd == NULL)
#ifdef MERGE1
	fprintf(fp,"Monodirectionality Retired (lc=%d)\n",leaf[DATA]->link_cnt);
#else
	fprintf(fp,"Monodirectionality Retired\n");
#endif MERGE1
    else
#endif RETIRE
    {
#ifdef MERGE1
	fprintf(fp,"Monodirectionality (lc=%d)\n",leaf[DATA]->link_cnt);
#else
	fprintf(fp,"Monodirectionality\n");
#endif MERGE1
	for (p = 0; p < parts; p++)
	{
	    fprintf(fp,"  Part %s dirs (%d) ",part_name(p),pd[p].ndir);
	    pr_dire_or(fp,&pd[p].d);
	    fprintf(fp,"\n");
	}    
	fprintf(fp,"%d Parts Decided using (%d) ",
	    leaf[DATA]->decided,leaf[DATA]->nused);
	pr_dire_and(fp,leaf[DATA]->used);
	fprintf(fp,"\nCommon Dirs for %d Uncovered parts: ",
	    leaf[DATA]->uncovered);
	pr_dire_and(fp,leaf[DATA]->others);
	fprintf(fp,"\n");
    }
}


/* MONO_DROP(leaf)
 * Release all memory used by our data structure in the given leaf.  If it is
 * a merged leaf, just decrement the link count.
 */

mono_drop(leaf)
Leaf leaf;
{
    if (is_on(SHOW_ENTRIES)) printf("mono_drop\n");
#ifdef MERGE1
	if (--leaf[DATA]->link_cnt == 0)
#endif MERGE1
	{
		mono_retire(leaf);
		m_free(leaf[DATA],sizeof(struct Leaf_Data));
	}
	leaf[DATA] = NULL;
}

mono_retire(leaf)
Leaf leaf;
{
	if (is_on(SHOW_ENTRIES)) printf("mono_retire\n");
	if (leaf[DATA]->dirs == NULL) return;
	m_free(leaf[DATA]->dirs,sizeof(struct pdirs) * part_number());
	freeset(leaf[DATA]->used);
	freeset(leaf[DATA]->others);
	leaf[DATA]->dirs = NULL;
}


/* MONO_COPY(new,old)
 * Copy the given leaf into a new one.  This is done by making both point to
 * the same leaf if MERGE1 is defined.  Otherwise a physical copy is always
 * made.
 *
 * MONO_SPLT(leaf)
 * This takes a leaf with a link count greater than one and divides it into
 * a separate leaf with it's own memory.  Used if MERGE1 is defined only.
 */

#ifdef MERGE1
mono_copy(new,old)
Leaf new,old;
{
	if (is_on(SHOW_ENTRIES)) printf("mono_copy\n");
	new[DATA] = old[DATA];
	new[DATA]->link_cnt++;
}

mono_split(new)
Leaf new;
{
struct Leaf_Data *od = new[DATA];
#else
mono_copy(new,old)
Leaf new,old;
{
struct Leaf_Data *od = old[DATA];
#endif MERGE1
struct pdirs *npd, *opd=od->dirs;
int parts = part_number();
Part p;

    if (is_on(SHOW_ENTRIES)) printf("mono_split\n");

    /* Allocate memory for new leaf */
    new[DATA] = (struct Leaf_Data *)m_alloc(sizeof(struct Leaf_Data));

    /* Copy data */
    new[DATA]->decided = od->decided;
    new[DATA]->nused = od->nused;
    new[DATA]->uncovered = od->uncovered;

#ifdef MERGE1
    new[DATA]->link_cnt = 1;
    od->link_cnt--;
#endif MERGE1
    
#ifdef RETIRE
    if (opd == NULL)
	new[DATA]->dirs = NULL;
    else
#endif RETIRE
    {
	/* Copy stuff not in retired leaves */

	new[DATA]->used = mkdireset();
	setcopy(new[DATA]->used,od->used);

	new[DATA]->others = mkdireset();
	if (od->uncovered != -1)
	    setcopy(new[DATA]->others,od->others);

	npd = new[DATA]->dirs =
	    (struct pdirs *)m_alloc(sizeof(struct pdirs) * parts);

	for (p=0; p<parts; p++)
	{
	    npd[p].ndir = opd[p].ndir;
	    setcopy(&npd[p].d,&opd[p].d);
	}
    }
} /* } */


#ifdef TEST_PROPOSALS

/* MONO_TEST(leaf,prop)
 * Return a (short sighted) rating of effect of applying proposal prop to a
 * leaf.  This rating should be the same as the one returned with the a
 * proposal by mono_prop.  If the proposal is better than it's negation,
 * a negative rating is returned.  If it is worse, a positive rating is
 * returned.  If both are the same, 0 is returned.
 */

int mono_test(leaf,prop)
Leaf leaf;
Prop *prop;
{
int parts = part_number();
struct pdirs *pd = leaf[DATA]->dirs;
Part p;
TrajSet ts;
TrajSet *ts1,*ts2;
DireSet ds1,ds2;
DireSet tds;

    if (is_on(SHOW_ENTRIES)) printf("test_prop\n");

    /* If we are retired, nothing matters */
#ifdef RETIRE
    if (pd == NULL) return(0);
#else
    if (leaf[DATA]->decided == parts) return(0);
#endif RETIRE

    /* If this is not a trajectory proposal, it is worthless */
    if (prop->type != TRAJ) return (0);

    /* CASE 1: No uncovered parts */
    if (leaf[DATA]->uncovered == 0)
    {
	/* Construct the set */
	p = prop->a.traj->p;
	ts1 = &prop->a.traj->t;
	traj_to_dir(&ds1,ts1);

	/* Does the proposed set of directions intersect the covering set? */
	if (isdisjoint(leaf[DATA]->used,&ds1))
		return(parts);

	/* Construct the set's negation */
	setdifference(ts2 = &ts,possible_traj(leaf,p),ts1);
	traj_to_dir(&ds2,ts2);

	/* Do the proposed sets of directions intersect the covering set? */
	if (isdisjoint(leaf[DATA]->used,&ds2))
		return(-parts);
	else
		return(0);
    }

    /* CASE 2: There is another direction that covers all uncovered parts */
    if (!isempty(leaf[DATA]->others))
    {
	setunion(&tds,leaf[DATA]->used,leaf[DATA]->others);

	/* Construct the set */
	p = prop->a.traj->p;
	ts1 = &prop->a.traj->t;
	traj_to_dir(&ds1,ts1);

	if (isdisjoint(&tds,&ds1))
		return(parts);

	/* Construct the set's negation */
	setdifference(ts2 = &ts,possible_traj(leaf,p),ts1);
	traj_to_dir(&ds2,ts2);

	if (isdisjoint(&tds,&ds2))
		return(-parts);
	else
		return(0);
    }

    /* CASE 3: Don't know what the minimal cover might be */
    return(0);
}
#else
int mono_test() {}
#endif TEST_PROPOSALS

#ifdef SIZE
mono_leaf_size(leaf,size,links)
Leaf leaf;
long *size;
int *links;
{
	*links = leaf[DATA]->link_cnt;
	*size = sizeof(struct Leaf_Data);
	if (leaf[DATA]->dirs != NULL)
		*size += (sizeof(struct pdirs) * part_number()) +
			 setmem(leaf[DATA]->used) +
			 setmem(leaf[DATA]->others);
}

long mono_stat_size()
{
    return (sizeof(Dire) * traj_number());
}
#endif SIZE
