/* SET.C -- Set Manipulation routines */
/* (c) 1988 - University of Michigan */

#include <stdio.h>
#include "config.h"
#include "global.h"
#include "set.h"
#ifndef NO_VARARGS
#include <varargs.h>
#endif NO_VARARGS

#define DATA SET_MOD

#ifndef NO_VARARGS

/* MKSET(maxsize,[elements],EOS)
 * Return a pointer to a newly allocated set of the given maximum size.
 * Initialize it by putting in all the elements listed in the args. The
 * list of elements must be terminated by an EOS.
 */

/*VARARGS*/ Set *mkset(va_alist)
va_dcl
{
va_list el;
register Set *s;
register int i;
int l,b;

	va_start(el);
	b = va_arg(el,int);
	l = llen(b);
	s = (Set *)m_alloc(setmeml(l));
	s->bsize = b;
	s->lsize = l;

	for (i=0; i < s->lsize; i++)
		s->elem.l[i] = 0L;

	while ((i=va_arg(el,int)) != EOS)
	{
#ifdef RANGECHECK
		if (i < 0 || i >= b)
			error("Element out of range in mkset\n");
#endif
		s->elem.b[i] = 1;
	}
	va_end(el);

	return(s);
}
#endif NO_VARARGS


/* FREESET
 * Release the memory allocated to a set created by mkset().
 */

freeset(s)
Set *s;
{
	m_free(s,setmem(s));
}

#ifndef NO_VARARGS

/* SETSET(set,maxsize,[elements],EOS)
 * Initialize a set with the given maximum size to the set of elements given.
 * The list of elements must be terminated by an EOS.
 */

/*VARARGS*/ setset(va_alist)
va_dcl
{
va_list el;
register Set *s;
register int i;

	va_start(el);
	s = (Set *)va_arg(el,Set *);
	s->bsize = va_arg(el,int);
	s->lsize = llen(s->bsize);
	for (i=0; i < s->lsize; i++)
		s->elem.l[i] = 0L;
	while ((i=va_arg(el,int)) != EOS)
	{
#ifdef RANGECHECK
		if (i < 0 || i >= s->bsize)
			error("Element out of range in setset\n");
#endif
		s->elem.b[i] = 1;
	}
	va_end(el);
}


/* APPENDN(set,[elements])
 * Add the listed elements to the named set.  The list of elements must be
 * terminated by EOS.  For one element, the macro append1() is faster.
 */

/*VARARGS*/ appendn(va_alist)
va_dcl
{
va_list el;
Set *s;
register int i;

	va_start(el);
	s=va_arg(el,Set *);
	while ((i=va_arg(el,int)) != EOS)
	{
#ifdef RANGECHECK
		if (i < 0 || i >= s->bsize)
			error("Element out of range in appendn\n");
#endif
		s->elem.b[i] = 1;
	}
	va_end(el);
}

/* REMOVEN(set,[elements])
 * Delete the listed elements from the named set.  The list of elements must
 * be terminated by EOS.  For one element, the macro remove1() is faster.
 */

/*VARARGS*/ removen(va_alist)
va_dcl
{
va_list el;
Set *s;
register int i;

	va_start(el);
	s=va_arg(el,Set *);
	while ((i=va_arg(el,int)) != EOS)
	{
#ifdef RANGECHECK
		if (i < 0 || i >= s->bsize)
			error("Element out of range in appendn\n");
#endif
		s->elem.b[i] = 0;
	}
	va_end(el);
}
#endif NO_VARARGS

/* SETCOPY(set,set)
 * The contents of the second set are copied to the first set.  They should
 * both be the same type.  If they were not, they will be after this
 * operation.
 */

setcopy(a,b)
Set *a,*b;
{
register int n;
register long *p=a->elem.l, *q=b->elem.l;

	a->bsize = b->bsize;
	a->lsize = n = b->lsize;

	while (n-- > 0)
		*(p++) = *(q++);
}

/* SETNOT(set,set)
 * The negation of the second set is copied to the first set.  They should
 * both be the same type.  They may both the the same set.  Note this is done
 * byte-by-byte, not long-by-long, because we don't want to invert extra
 * memory in sets whose size is not a multiple of four.
 */

setnot(a,b)
Set *a,*b;
{
register int n;
register char *p=a->elem.b, *q=b->elem.b;

	a->bsize = n = b->bsize;
	a->lsize = b->lsize;

	while (n-- > 0)
		*(p++) = ~(*(q++));
}

/* SETDIFFERENCE(setc,seta,setb)
 * The difference a-b is copied into set c.  All sets must be the same type.
 * Set c may be the same set as either of the other two.
 */

setdifference(c,a,b)
Set *a,*b,*c;
{
register int n;
register long *p=a->elem.l, *q=b->elem.l, *r=c->elem.l;

	c->bsize = a->bsize;
	c->lsize = n = a->lsize;

	while (n-- > 0)
		*(r++) = *(p++) & ~(*(q++));
}

/* SETUNION(setc,seta,setb)
 * The union a|b is copied into set c.  All sets must be the same type.
 * Set c may be the same set as either of the other two.
 */

setunion(c,a,b)
Set *a,*b,*c;
{
register int n;
register long *p=a->elem.l, *q=b->elem.l, *r=c->elem.l;

	c->bsize = a->bsize;
	c->lsize = n = a->lsize;

	while (n-- > 0)
		*(r++) = *(p++) | *(q++);
}

/* SETINTERSECTION(setc,seta,setb)
 * The intersection a&b is copied into set c.  All sets must be the same type.
 * Set c may be the same set as either of the other two.  Returns TRUE
 * if result is not empty.
 */

boolean setintersection(c,a,b)
Set *a,*b,*c;
{
register long *p=a->elem.l, *q=b->elem.l, *r=c->elem.l;
register int n;
register long flag=0;

	c->bsize = a->bsize;
	c->lsize = n = a->lsize;

	while (n-- > 0)
		flag |= (*(r++) = *(p++) & *(q++));
	return(flag != 0);
}

/* ISDISJOINT(seta,setb)
 * Return FALSE if set a contains any element in set b.  Both sets must
 * be of the same type.
 */

boolean isdisjoint(a,b)
Set *a,*b;
{
register int n=a->lsize;
register long *p=a->elem.l, *q=b->elem.l;

	while (n-- > 0)
		if (*(p++) & *(q++)) return(FALSE);
	return(TRUE);
}


/* ISSUBSET(seta,setb)
 * Return FALSE if set a contains any element not in set b.  Both sets must
 * be of the same type.
 */

boolean issubset(a,b)
Set *a,*b;
{
register int n=a->lsize;
register long *p=a->elem.l, *q=b->elem.l;

	while (n-- > 0)
		if (*(p++) & ~(*(q++))) return(FALSE);
	return(TRUE);
}

/* ISEQUAL(seta,setb)
 * Return FALSE if set's a and b are equal.  Both sets must be of the same
 * type.
 */

boolean isequal(a,b)
Set *a,*b;
{
register int n=a->lsize;
register long *p=a->elem.l, *q=b->elem.l;

	while (n-- > 0)
		if (*(p++) != *(q++)) return(FALSE);
	return(TRUE);
}

/* ISEMPTY(set)
 * Return TRUE if set is empty.
 */

boolean isempty(a)
Set *a;
{
register int n=a->lsize;
register long *p=a->elem.l;

	while (n-- > 0)
		if (*(p++)) return(FALSE);
	return(TRUE);
}

/* PR_SET(fp,s,l,namefun)
 * Print the contents of a set to fp, calling namefun() to get an element name.
 * There are a bunch of macros in set.h that are normally used with this.
 */

pr_set(fp,s,l,name)
FILE *fp;
Set *s;
char *l;
char *((*name)());
{
char *a=NULL, *b=NULL;
Part p;

	foreach(p,s)
	{
		if (b) fprintf(fp,"%s, ",b);
		b = a;
		a = (*name)(p);
	}
	if (b)
		fprintf(fp,"%s%s ",b,l);
	if (a)
		fprintf(fp,"%s",a);
	else
		fprintf(fp,"nothing");
}

dumpset(s)
Set *s;
{
int i;
	for (i=0;i<s->bsize;i++)
		putc(s->elem.b[i] ? '1': '0',stderr);
	putc('\n',stderr);
}

/* CARDINALITY(s)
 * Return the number of elements in a set
 */

int cardinality(s)
register Set *s;
{
register int n = 0;
register int i;

	foreach(i,s)
		n++;
	return(n);
}


/* FIRST_IN(s)
 * Return the first element in a set.
 */

int first_in(s)
register Set *s;
{
register int i;

	foreach(i,s)
		return(i);
}


/* LAST_IN(s)
 * Return the last element in a set.
 */

int last_in(s)
register Set *s;
{
register int i;

	for (i=s->bsize-1; i>=0; i--)
		if (isin(s,i)) return(i);
}

#ifdef NO_VARARGS

/* MKUSET(maxsize,element)
 * Return a pointer to a newly allocated set of the given maximum size.
 * Initialize it by putting the element in the args. If it is to be
 * initialized empty, the element must be an EOS.
 */

Set *mkuset(b,e)
int b,e;
{
register Set *s;
register int i;
int l;

	l = llen(b);
	s = (Set *)m_alloc(setmeml(l));
	s->bsize = b;
	s->lsize = l;

	for (i=0; i < s->lsize; i++)
		s->elem.l[i] = 0L;

	if (e != EOS)
	{
#ifdef RANGECHECK
		if (e < 0 || e >= b)
			error("Element out of range in mkuset\n");
#endif
		s->elem.b[e] = 1;
	}
	return(s);
}

/* SETUSET(set,maxsize,element)
 * Initialize a set with the given maximum size to contain the named element.
 * To empty the set, the element must be an EOS.
 */

setuset(s,b,e)
Set *s;
int b,e;
{
register int i;

	s->bsize = b;
	s->lsize = llen(b);
	for (i=0; i < s->lsize; i++)
		s->elem.l[i] = 0L;
	if (e != EOS)
	{
#ifdef RANGECHECK
		if (e < 0 || e >= s->bsize)
			error("Element out of range in setuset\n");
#endif
		s->elem.b[e] = 1;
	}
}
#endif NO_VARARGS
