/* SET.H */
/* (c) 1988 - University of Michigan */

#ifndef _H_SET
#define _H_SET

#ifndef _H_CONFIG
#include "config.h"
#endif _H_CONFIG

/* LLEN(n)
 * The number of longs needed that cover an array of n characters.
 */

#ifdef LONG32
#   define llen(n) ((n+3)>>2)
#else
#   define llen(n) ((n+sizeof(long)-1)/sizeof(long))
#endif

/* DEFSET(x,n)
 * Generate a typedef with name x for a set containing at most n elements.
 * Sets are simply arrays of characters, each of which is one if the coor-
 * isponding element is in the set, and zero if it is not.  A size is attached
 * to the structure so we can write genearic functions to handle them.  They
 * are also addressable as an array of longs, for faster union/intersection
 * type operations.  Individual set elements are represented by their indexes.
 */

#define defset(x,n) typedef struct {setsize bsize, lsize; union {char b[n]; long l[llen(n)];} elem;} x


/* PARTSET, TRAJSET, SET
 * Two set types for parts and trajectories respectively, and a phony set type
 * that gets used when we don't know or care which of the others we are using.
 */

defset(PartSet,MAXPARTS);
defset(TrajSet,MAXTRAJS);
typedef TrajSet Set;

/* EOS
 * "End of Set" is used to mark the end of a list of set elements given in
 * a procedure call.
 */

#define EOS -1

/* SETMEM(s)
 * Return the actual amount of memory used for a given set from i.
 */
#define setmeml(l) sizeof(TrajSet) - sizeof(long)*(llen(MAXTRAJS) - (l))
#define setmem(s) setmeml((s->lsize))

/* MAKESET and INITSET
 * These are like setset and mkset, for sets with 0 or 1 element in them
 * initially, which is all we happen to ever use, which is a good thing
 * since with NO_VARARGS larger numbers don't work anyway.
 */

#ifdef NO_VARARGS
#  define makeset mkuset
#  define initset setuset
#else
#  define makeset mkset
#  define initset setset
#endif NO_VARARGS

/* MKPARTSET, MKTRAJSET
 * These routines allocate an empty set of the named type.
 * EMPTYPARTSET, EMPTYTRAJSET
 * These routines initialize an empty set of the named type.
 * SETPARTSET, SETTRAJSET
 * These routines initialize an single element set of the named type.
 */
#define mkpartset() ((PartSet *) makeset(part_number(),EOS))
#define mktrajset() ((TrajSet *) makeset(traj_number(),EOS))
#define emptypartset(s) initset(s,part_number(),EOS)
#define emptytrajset(s) initset(s,traj_number(),EOS)
#define setpartset(s,e) initset(s,part_number(),e,EOS)
#define settrajset(s,e) initset(s,traj_number(),e,EOS)

/* APPEND1(set,element)
 * REMOVE1(set,element)
 * Add or delete the given element from the given set.  For multiple elements
 * the functions appendn() and removen() can be used.
 */

#ifdef RANGECHECK
#  define append1(s,e) if (e < 0 || e >= (s)->bsize) error("element out of range in append1 macro\n"); else (s)->elem.b[e] = 1
#  define remove1(s,e) if (e < 0 || e >= (s)->bsize) error("element out of range in remove1 macro\n"); else (s)->elem.b[e] = 0
#else
#  define append1(s,e) ((s)->elem.b[e] = 1)
#  define remove1(s,e) ((s)->elem.b[e] = 0)
#endif

/* REMOVE(set,set)
 * APPEND(set,set)
 * The contents of the second set are added or deleted from the first set.
 * They should both be the same type.  One should be darn careful about
 * any side effects of the first arg.
 */

#define append(a,b) setunion(a,a,b)
#define remove(a,b) setdifference(a,a,b)
#define intersect(a,b) setintersection(a,a,b)

/* NEGATE(set)
 * Negate the given set.  There should be no side effects.
 */

#define negate(s) setnot(s,s)

/* ISIN(set,element)
 * Return TRUE if the given set contains the given element.  This doesn't
 * check if the element is in range.
 */

#ifdef RANGECHECK
#  define isin(s,e) (((e) >= 0 && (e) < (s)->bsize) ? (s)->elem.b[e] : (error("element out of range in isin macro\n")))
#else
#  define isin(s,e) ((s)->elem.b[e])
#endif

/* FOREACH(element,set)
 * This is a loop statement.  It repeats with the variable element set to
 * each element in set successively.
 */

#ifdef RANGECHECK
#define foreach(e,s) for (e=0; e < (s)->bsize; e++) if ((s)->elem.b[e])
#else
#define foreach(e,s) for (e=0; e < (s)->bsize; e++) if (isin(s,e))
#endif

/* PRINT_SET macros.
 * Print a set of a certain type with either an "and" or an "or".
 */

#define pr_set_and(f,s,n) pr_set(f,(s)," and",n);
#define pr_set_or(f,s,n) pr_set(f,(s)," or",n);
#define pr_set_comma(f,s,n) pr_set(f,(s),",",n);

#define pr_part_and(f,s) pr_set_and(f,s,part_name)
#define pr_traj_and(f,s) pr_set_and(f,s,traj_name)
#define pr_dire_and(f,s) pr_set_and(f,s,dire_name)

#define pr_part_or(f,s) pr_set_or(f,s,part_name)
#define pr_traj_or(f,s) pr_set_or(f,s,traj_name)
#define pr_dire_or(f,s) pr_set_or(f,s,dire_name)

#define pr_part_comma(f,s) pr_set_comma(f,s,part_name)
#define pr_traj_comma(f,s) pr_set_comma(f,s,traj_name)
#define pr_dire_comma(f,s) pr_set_comma(f,s,dire_name)

/* declarations */
Set *mkset();

#endif _H_SET
